import os
import hashlib
import json
import tempfile
import collections

from unittest import mock
from click.testing import CliRunner

from hamlet.command.layer import describe_layer_type, describe_layer
from hamlet.command.common.config import Options
from tests.unit.command.test_option_generation import (
    run_options_test,
)


info_mock_output = {
    "LayerTypes": [
        {
            "Type": "LayerType[1]",
            "Description": "LayerDescription[1]",
            "Attributes": [{"Names": "Attribute[1]", "Types": ["string"]}],
            "ReferenceLookupType": "LayerReferenceLookupType[1]",
        },
        {
            "Type": "LayerType[2]",
            "Description": "LayerDescription[2]",
            "Attributes": [{"Names": "Attribute[2]", "Types": ["string"]}],
            "ReferenceLookupType": "LayerReferenceLookupType[2]",
        },
    ],
    "LayerData": [
        {
            "Type": "LayerType[1]",
            "Id": "LayerDataId[1]",
            "Name": "LayerDataName[1]",
            "Properties": {"LayerProperty[1]": "LayerPropertyValue[1]"},
        },
        {
            "Type": "LayerType[2]",
            "Id": "LayerDataId[2]",
            "Name": "LayerDataName[2]",
            "Properties": {"LayerProperty[2]": "LayerPropertyValue[2]"},
        },
    ],
}


def template_backend_run_mock(data):
    def run(
        entrance="info",
        output_filename="info.json",
        deployment_mode=None,
        output_dir=None,
        generation_input_source=None,
        generation_provider=None,
        generation_framework=None,
        log_level=None,
        root_dir=None,
        tenant=None,
        account=None,
        product=None,
        environment=None,
        segment=None,
    ):
        os.makedirs(output_dir, exist_ok=True)
        filename = os.path.join(output_dir, output_filename)
        with open(filename, "wt+") as f:
            json.dump(data, f)

    return run


def mock_backend(data=None):
    def decorator(func):
        @mock.patch("hamlet.backend.query.context.Context")
        @mock.patch("hamlet.backend.query.template")
        def wrapper(blueprint_mock, ContextClassMock, *args, **kwargs):
            with tempfile.TemporaryDirectory() as temp_cache_dir:

                ContextObjectMock = ContextClassMock()
                ContextObjectMock.md5_hash.return_value = str(
                    hashlib.md5(str(data).encode()).hexdigest()
                )
                ContextObjectMock.cache_dir = temp_cache_dir
                blueprint_mock.run.side_effect = template_backend_run_mock(data)

                return func(blueprint_mock, ContextClassMock, *args, **kwargs)

        return wrapper

    return decorator


DECSCRIBE_TYPE_VALID_OPTIONS = collections.OrderedDict()
DECSCRIBE_TYPE_VALID_OPTIONS["!-t,--type"] = "ReferenceType[1]"
DECSCRIBE_TYPE_VALID_OPTIONS["-q,--query"] = "[]"


@mock_backend(info_mock_output)
def test_describe_layer_type_input_valid(
    blueprint_mock,
    ContextClassMock,
):
    run_options_test(
        CliRunner(),
        describe_layer_type,
        DECSCRIBE_TYPE_VALID_OPTIONS,
        blueprint_mock.run,
    )


@mock_backend(info_mock_output)
def test_describe_layer_type(blueprint_mock, ContextClassMock):
    obj = Options()

    cli = CliRunner()
    result = cli.invoke(describe_layer_type, ["--type", "LayerType[1]"], obj=obj)
    print(result.exc_info)
    assert result.exit_code == 0
    result = json.loads(result.output)
    assert {
        "Type": "LayerType[1]",
        "Description": "LayerDescription[1]",
        "Attributes": [{"Names": "Attribute[1]", "Types": ["string"]}],
        "ReferenceLookupType": "LayerReferenceLookupType[1]",
    } == result


DECSCRIBE_LAYER_VALID_OPTIONS = collections.OrderedDict()
DECSCRIBE_LAYER_VALID_OPTIONS["!-t,--type"] = "LayerType[1]"
DECSCRIBE_LAYER_VALID_OPTIONS["!-n,--name"] = "LayerDataName[1]"
DECSCRIBE_LAYER_VALID_OPTIONS["-q,--query"] = "[]"


@mock_backend(info_mock_output)
def test_describe_layer_input_valid(
    blueprint_mock,
    ContextClassMock,
):
    run_options_test(
        CliRunner(),
        describe_layer,
        DECSCRIBE_LAYER_VALID_OPTIONS,
        blueprint_mock.run,
    )


@mock_backend(info_mock_output)
def test_describe_layer(blueprint_mock, ContextClassMock):
    obj = Options()

    cli = CliRunner()
    result = cli.invoke(
        describe_layer,
        ["--type", "LayerType[1]", "--name", "LayerDataName[1]"],
        obj=obj,
    )
    print(result.exc_info)
    assert result.exit_code == 0
    result = json.loads(result.output)
    assert {
        "Type": "LayerType[1]",
        "Id": "LayerDataId[1]",
        "Name": "LayerDataName[1]",
        "Properties": {"LayerProperty[1]": "LayerPropertyValue[1]"},
    } == result



import click
from hamlet.backend.manage import file_crypto as manage_file_crypto_backend
from hamlet.command.common.config import pass_options
from hamlet.command.common import exceptions


@click.command(
    "file-crypto",
    short_help="Manage crypto for files",
    context_settings=dict(max_content_width=240),
)
@click.option("-d", "--decrypt", is_flag=True, help="decrypt file")
@click.option("-e", "--encrypt", is_flag=True, help="encrypt file")
@click.option(
    "-f",
    "--crypto-file",
    help="path to the file managed",
    type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True),
)
@click.option("-u", "--update", is_flag=True, help="update the file")
@exceptions.backend_handler()
@pass_options
def file_crypto(options, **kwargs):
    """
    Manage crypto for files

    \b
    NOTES:
    1. If no operation is provided, the current file contents are displayed
    """

    args = {**options.opts, **kwargs}

    manage_file_crypto_backend.run(**args, _is_cli=True)



import os
import tempfile
import click

from hamlet.command.common import exceptions
from hamlet.command.common.config import pass_options
from hamlet.command.common.exceptions import CommandError
from hamlet.backend.create import template as create_template_backend
from hamlet.backend.test.generate import run as test_generate_backend
from hamlet.backend.test import run as test_run_backend
from hamlet.backend.deploy import find_deployments, create_deployment


@click.command(
    "test-deployments", short_help="", context_settings=dict(max_content_width=240)
)
@click.option(
    "-m",
    "--deployment-mode",
    default="update",
    help="The deployment mode to use for the deployment",
)
@click.option(
    "-l",
    "--deployment-group",
    default=".*",
    show_default=True,
    help="The deployment group pattern to match",
)
@click.option(
    "-u",
    "--deployment-unit",
    default=[".*"],
    show_default=True,
    multiple=True,
    help="The deployment unit pattern to match",
)
@click.option(
    "-d",
    "--district",
    default=["segment"],
    multiple=True,
    help="The districts to include deployments from",
)
@click.option(
    "-o",
    "--output-dir",
    type=click.Path(
        file_okay=False,
        dir_okay=True,
        writable=True,
        readable=True,
    ),
    help="The directory where tests are run - temp dir by default",
)
@click.option(
    "-p", "--pytest-args", "pytest_args", help="additional arguments for pytest"
)
@click.option("-s", "--silent", "silent", is_flag=True, help="minimize pytest output")
@exceptions.backend_handler()
@pass_options
def test_deployments(
    options,
    deployment_mode,
    deployment_group,
    deployment_unit,
    district,
    output_dir,
    pytest_args,
    silent,
    **kwargs,
):
    """
    Create deployments and test them using their test cases
    """
    temp_dir = None

    if output_dir is None:
        temp_dir = tempfile.TemporaryDirectory()
        output_dir = temp_dir.name

    output_dir = os.path.abspath(output_dir)

    deployments = find_deployments(
        deployment_mode=deployment_mode,
        deployment_group=deployment_group,
        deployment_units=deployment_unit,
        districts=district,
        **options.opts,
    )

    if len(deployments) == 0:
        raise exceptions.CommandError("No deployments found that match pattern")

    click.echo("")
    click.echo(click.style("[*] Creating deployments:", bold=True, fg="green"))
    click.echo("")
    for deployment in deployments:

        deployment_group = deployment["DeploymentGroup"]
        deployment_unit = deployment["DeploymentUnit"]

        click.echo(f"[-] {deployment_group}/{deployment_unit}")

        create_deployment(
            deployment_group,
            deployment_unit,
            deployment_mode,
            output_dir,
            _is_cli=False,
            **options.opts,
        )

        generate_args = {
            **options.opts,
            "entrance": "deploymenttest",
            "deployment_group": deployment_group,
            "deployment_unit": deployment_unit,
            "deployment_mode": deployment_mode,
            "output_dir": output_dir,
        }
        create_template_backend.run(**generate_args, _is_cli=False)

    click.echo("")
    click.secho("[*] Testing deployments:", bold=True, fg="green")
    click.echo("")

    test_script_filename = "test_deployments.py"
    test_generate_backend(
        output=f"{output_dir}/{test_script_filename}",
        directory=output_dir,
    )

    try:
        test_run_backend.run(
            testpaths=[test_script_filename],
            pytest_args=pytest_args,
            silent=silent,
            root_dir=output_dir,
        )
    except Exception as e:
        raise CommandError(str(e))



from hamlet.backend.common import runner


def run(
    generation_input_source=None,
    output_dir=None,
    generation_provider=None,
    generation_framework=None,
    generation_testcase=None,
    log_level=None,
    root_dir=None,
    tenant=None,
    account=None,
    product=None,
    environment=None,
    segment=None,
    _is_cli=False,
):
    options = {
        "-i": generation_input_source,
        "-o": output_dir,
        "-p": generation_provider,
        "-f": generation_framework,
        "-t": generation_testcase,
        "-l": "unitlist",
    }
    env = {
        "GENERATION_LOG_LEVEL": log_level,
        "ROOT_DIR": root_dir,
        "TENANT": tenant,
        "ACCOUNT": account,
        "PRODUCT": product,
        "ENVIRONMENT": environment,
        "SEGMENT": segment,
    }
    runner.run("createTemplate.sh", [], options, env, _is_cli)



from hamlet.backend.common import runner


def run(
    deployment_unit=None,
    run_setup=None,
    force_binary_build=None,
    submit_binary=None,
    binary_build_process=None,
    build_logs=None,
    environment_badge=None,
    environment_badge_content=None,
    node_package_manager=None,
    app_version_source=None,
    log_level=None,
    root_dir=None,
    tenant=None,
    account=None,
    product=None,
    environment=None,
    segment=None,
    _is_cli=False,
    **kwargs
):
    options = {
        "-b": binary_build_process,
        "-d": environment_badge_content,
        "-e": environment_badge,
        "-f": force_binary_build,
        "-l": build_logs,
        "-m": submit_binary,
        "-n": node_package_manager,
        "-s": run_setup,
        "-u": deployment_unit,
        "-v": app_version_source,
    }
    env = {
        "GENERATION_LOG_LEVEL": log_level,
        "ROOT_DIR": root_dir,
        "TENANT": tenant,
        "ACCOUNT": account,
        "PRODUCT": product,
        "ENVIRONMENT": environment,
        "SEGMENT": segment,
    }
    runner.run("runExpoAppPublish.sh", [], options, env, _is_cli)



import os
import base64
import tempfile
import hashlib
import shutil
import httpx
import typing
import www_authenticate
import json

from urllib import parse


class ContainerRepositoryException(BaseException):
    def __init__(self, repository: str, *args: object) -> None:
        self.repository = repository
        super().__init__(*args)


class ContainerTagNotFoundException(ContainerRepositoryException):
    def __init__(self, repository: str, tag: str, *args: object) -> None:
        self.tag = tag
        super().__init__(repository, *args)


class DockerRegistryV2Auth(httpx.Auth):
    def __init__(self, repository, username=None, password=None) -> None:
        self.repository = repository
        self.username = username
        self.password = password
        self.token = None

    actions = ["pull"]

    def auth_flow(
        self, request: httpx.Request
    ) -> typing.Generator[httpx.Request, httpx.Response, None]:

        if self.token:
            request.headers["Authorization"] = self.token

        response = yield request

        if response.status_code == httpx.codes.UNAUTHORIZED:
            self.token = self.sign_request(
                response, self.repository, self.actions, self.username, self.password
            )
            request.headers["Authorization"] = self.token

            yield request

    def sign_request(
        self, registry_response, repository, actions, username, password
    ) -> str:
        realm_client = httpx.Client()

        if self.username or self.password:
            realm_client.headers["Authorization"] = self._build_basic_auth_header(
                username, password
            )

        realm_response = realm_client.get(
            self._build_realm_token_url(registry_response, repository, actions)
        )

        try:
            realm_response.raise_for_status()
        except httpx.HTTPError as e:
            raise e

        realm_token = json.loads(realm_response.text).get("access_token") or json.loads(
            realm_response.text
        ).get("token")
        return f"Bearer {realm_token}"

    def _build_basic_auth_header(self, username, password) -> str:

        userpass = b":".join((httpx.to_bytes(username), httpx.to_bytes(password)))
        token = base64.b64encode(userpass).decode()
        return f"Basic {token}"

    def _build_realm_token_url(self, response, repository, actions) -> str:
        www_authenticate_response = www_authenticate.parse(
            response.headers["WWW-Authenticate"]
        )

        if "bearer" in www_authenticate_response:
            if self.actions:
                scope = f"repository:{repository}:" + ",".join(actions)
            elif "scope" in www_authenticate_response["bearer"]:
                scope = www_authenticate_response["bearer"]["scope"]
            else:
                scope = ""

            url_parts = list(
                parse.urlparse(www_authenticate_response["bearer"]["realm"])
            )
            query = parse.parse_qs(url_parts[4])
            query.update(
                {
                    "service": www_authenticate_response["bearer"]["service"],
                    "scope": scope,
                }
            )
            url_parts[4] = parse.urlencode(query, True)
            url_parts[0] = "https"

            return parse.urlunparse(url_parts)


class ContainerRepository:
    def __init__(self, registry_url, repository, username=None, password=None) -> None:

        self.registry_url = registry_url
        self.repository = repository

        self.username = username
        self.password = password

        self.registry_client = httpx.Client(
            base_url=self.registry_url,
            auth=DockerRegistryV2Auth(self.repository, self.username, self.password),
            headers={"Accept": "application/vnd.docker.distribution.manifest.v2+json"},
        )

    @property
    def tags(self):
        tag_response = self.registry_client.get(
            f"/v2/{self.repository}/tags/list", allow_redirects=True
        )
        try:
            tag_response.raise_for_status()
        except httpx.HTTPError as e:
            raise e

        return json.loads(tag_response.text).get("tags")

    def get_tag_digest(self, tag):
        return self.get_tag_manifest(tag).headers["docker-content-digest"]

    def get_tag_manifest(self, tag):
        manifest_response = self.registry_client.get(
            f"/v2/{self.repository}/manifests/{tag}", allow_redirects=True
        )

        try:
            manifest_response.raise_for_status()

        except httpx.HTTPStatusError as e:
            if e.response.status_code == httpx.codes.NOT_FOUND:
                raise ContainerTagNotFoundException(self.repository, tag, e)

            raise e

        return manifest_response

    def pull(self, tag, dst_dir):
        manifest_response = self.get_tag_manifest(tag)
        manifest = manifest_response.json()

        with tempfile.TemporaryDirectory() as extract_dir:
            with tempfile.TemporaryDirectory() as stage_dir:
                for layer in manifest["layers"]:
                    with tempfile.NamedTemporaryFile(
                        dir=stage_dir, delete=False
                    ) as layer_file:

                        layer_blob = self.registry_client.get(
                            url=f'/v2/{self.repository}/blobs/{layer["digest"]}',
                            allow_redirects=True,
                        )

                        try:
                            layer_blob.raise_for_status()
                        except httpx.HTTPError as e:
                            raise e

                        with self.registry_client.stream(
                            method="get",
                            url=layer_blob.url,
                        ) as r:
                            for chunk in r.iter_raw(chunk_size=(1024 * 1024 * 3)):
                                layer_file.write(chunk)

                        layer_file.flush()

                        digest_algorithm = layer["digest"].split(":")[0]
                        digest_hash = layer["digest"].split(":")[1]

                        file_hash = hashlib.new(digest_algorithm)
                        with open(layer_file.name, "rb") as f:
                            for byte_block in iter(
                                lambda: f.read((1024 * 1024 * 10)), b""
                            ):
                                file_hash.update(byte_block)

                        if digest_hash != file_hash.hexdigest():
                            print(
                                (
                                    "Layer hash does not match digest "
                                    f"- ours {file_hash.hexdigest()} - theirs {digest_hash}"
                                )
                            )

                        if (
                            layer["mediaType"]
                            == "application/vnd.docker.image.rootfs.diff.tar.gzip"
                        ):

                            shutil.unpack_archive(layer_file.name, extract_dir, "gztar")

                if os.listdir(extract_dir):
                    if os.path.isdir(dst_dir):
                        shutil.rmtree(dst_dir)
                    shutil.copytree(extract_dir, dst_dir, symlinks=True)

        return manifest_response.headers["docker-content-digest"]



import os
import json
from unittest import mock

from hamlet.backend.engine.common import ENGINE_GLOBAL_NAME
from hamlet.backend.deploy import create_deployment as create_deployment_backend
from .conftest import conf


TENANT = conf["cmdb"]["tenant"]["tenant_name"]
ACCOUNT = conf["cmdb"]["account"]["account_name"]
PRODUCT = conf["cmdb"]["product"]["product_name"]
ENVIRONMENT = conf["cmdb"]["product"]["environment_name"]
SEGMENT = conf["cmdb"]["product"]["segment_name"]

LEVEL = "segment"


#  test is made accordingly to quickstart guide
@mock.patch.dict(
    os.environ,
    {
        "TENANT": TENANT,
        "ACCOUNT": ACCOUNT,
        "PRODUCT": PRODUCT,
        "ENVIRONMENT": ENVIRONMENT,
        "SEGMENT": SEGMENT,
    },
)
def run(cmdb, engine):
    global_engine_env = engine.engine_store.get_engine(ENGINE_GLOBAL_NAME).environment
    with mock.patch.dict(os.environ, {"ROOT_DIR": cmdb.ROOT_DIR, **global_engine_env}):

        create_deployment_backend(
            deployment_group=LEVEL,
            deployment_unit="baseline",
            deployment_mode="update",
            output_dir=None,
        )

        #  check that files were created
        with cmdb(
            PRODUCT, "infrastructure", "cf", ENVIRONMENT, SEGMENT, "baseline", "default"
        ) as filename:

            def prefix(unit):

                level_prefix = LEVEL
                if LEVEL == "segment":
                    level_prefix = "seg"

                return [
                    level_prefix,
                    unit,
                    ACCOUNT,
                    conf["cmdb"]["tenant"]["default_region"],
                ]

            assert os.path.exists(
                filename(*prefix("baseline"), "epilogue", sep="-", ext="sh")
            )

            assert os.path.exists(
                filename(
                    *prefix("baseline"), "generation-contract", sep="-", ext="json"
                )
            )

            cf_baseline_template_filename = filename(
                *prefix("baseline"), "template", sep="-", ext="json"
            )

            assert os.path.exists(cf_baseline_template_filename)
            with open(cf_baseline_template_filename) as f:
                json.load(f)



class JSONValidator:
    def __init__(self, body):
        self._body = body
        self._validators = []

    @classmethod
    def from_file(cls, filename):
        try:
            with open(filename, "rt") as f:
                import json

                return cls(json.load(f))
        except FileNotFoundError as e:
            raise AssertionError("%s not found" % filename) from e
        except ValueError as e:
            raise AssertionError("%s is not a valid JSON" % filename) from e

    @staticmethod
    def _split_path(path):
        import re

        keys = []
        for m in re.finditer(r"(\w+)|\[(\d+)\]", path):
            key = m.group(1)
            index = m.group(2)
            if key is not None:
                keys.append(key)
            elif index is not None:
                keys.append(int(index))
        return keys

    @staticmethod
    def _format_keys_to_path(keys):
        result = ""
        for key in keys:
            if isinstance(key, int):
                result += "[%s]" % key
            else:
                result += ".%s" % key
        if result.startswith("."):
            result = result[1:]
        return result

    def _get_value_by_json_path(self, path):
        keys = self._split_path(path)
        traversed_keys = []
        value = self._body
        for key in keys:
            traversed_keys.append(key)
            try:
                value = value[key]
            except (KeyError, IndexError) as e:
                raise AssertionError(
                    "{} does not exist".format(
                        self._format_keys_to_path(traversed_keys)
                    )
                ) from e
        return value

    @property
    def errors(self):
        try:
            return self._errors
        except AttributeError:
            errors = []
            for validator in self._validators:
                try:
                    validator()
                except AssertionError as e:
                    msg = None
                    if e.args:
                        msg = e.args[0].split("\n")[0]
                    errors.append({"rule": validator.__name__, "msg": msg})
            self._errors = errors
            return errors

    def assert_structure(self):
        import json

        if self.errors:
            raise AssertionError(json.dumps(self.errors, indent=4))



import os
import shutil
import pytest
from hamlet.loggers import logging, root
from hamlet.backend.engine import EngineStore


root.setLevel(logging.DEBUG)


class __CMDBContext:

    if os.getenv("TEST_ROOT_DIR", None) is not None:
        ROOT_DIR = os.getenv("TEST_ROOT_DIR")
    else:
        ROOT_DIR = os.getcwd() + "/.cmdb"
        os.makedirs(ROOT_DIR, exist_ok=True)

    def create_cmdb_filename_compositor(self):
        root = self.dir

        def cmdb_filename_compositor(*args, sep="-", ext=""):
            name = sep.join([*args])
            filename = os.path.join(root, name)
            if ext:
                filename += ".%s" % ext
            return os.path.normpath(filename)

        return cmdb_filename_compositor

    def __init__(self, *args):
        self.dir = os.path.join(self.ROOT_DIR, *args)
        self.old_cwd = os.getcwd()

    def __enter__(self):
        assert os.path.exists(self.dir), "Path %s does not exit" % self.dir
        assert os.path.isdir(self.dir), "Path %s is not a dir" % self.dir
        os.chdir(self.dir)
        return self.create_cmdb_filename_compositor()

    def __exit__(self, exc_type, exc_val, exc_tb):
        os.chdir(self.old_cwd)


def __clear_dir(dir):
    assert os.path.isdir(dir)
    for entry in os.listdir(dir):
        entry = os.path.abspath(os.path.join(dir, entry))
        if os.path.isdir(entry):
            shutil.rmtree(entry)
        else:
            os.remove(entry)


def __clear_cmdb(*args):
    with __CMDBContext() as path:
        if not args:
            __clear_dir(path())
        else:
            __clear_dir(os.path.join(*args))


@pytest.fixture(scope="session")
def cmdb():
    return __CMDBContext


@pytest.fixture(scope="session")
def clear_cmdb():
    return __clear_cmdb


class __EngineContext:

    if os.getenv("TEST_ENGINE_STORE_DIR", None) is not None:
        ENGINE_STORE_DIR = os.getenv("TEST_ENGINE_STORE_DIR")
    else:
        ENGINE_STORE_DIR = os.getcwd() + "/.engine_store"
        os.makedirs(ENGINE_STORE_DIR, exist_ok=True)

    engine_store = EngineStore(store_dir=ENGINE_STORE_DIR)


@pytest.fixture(scope="session")
def engine():
    return __EngineContext


def __clear_engine(*args):
    if os.path.exists(__EngineContext.ENGINE_STORE_DIR) and os.path.isdir(
        __EngineContext.ENGINE_STORE_DIR
    ):
        shutil.rmtree(__EngineContext.ENGINE_STORE_DIR)


@pytest.fixture(scope="session")
def clear_engine():
    return __clear_engine



import click
import os
import json

from tabulate import tabulate

from hamlet.command import root as cli
from hamlet.command.common import exceptions, config
from hamlet.command.common.display import json_or_table_option, wrap_text
from hamlet.command.common.config import pass_options

from hamlet.backend.engine import engine_store
from hamlet.backend.engine.common import ENGINE_GLOBAL_NAME
from hamlet.backend.engine.engine_code_source import EngineCodeSourceBuildData
from hamlet.backend.engine.exceptions import (
    HamletEngineInvalidVersion,
)


def engines_table(data):
    tablerows = []
    for row in data:
        tablerows.append(
            [
                wrap_text(row["name"]),
                wrap_text(row["description"]),
                wrap_text(row["installed"]),
                wrap_text(row["global"]),
                wrap_text(row["update_available"]),
            ]
        )
    return tabulate(
        tablerows,
        headers=["Name", "Description", "Installed", "Global", "Update Available"],
        tablefmt="github",
    )


@cli.group("engine", context_settings=dict(max_content_width=240))
def group():
    """
    Manage the engine used by the executor
    """


@group.command(
    "list-engines", short_help="", context_settings=dict(max_content_width=240)
)
@click.option(
    "--show-hidden/--hide-hidden",
    default=False,
    help="Control visibility of hidden engines",
)
@json_or_table_option(engines_table)
@exceptions.backend_handler()
def list_engines(show_hidden):
    """
    Lists the available engines
    """
    data = []

    for engine in engine_store.find_engines(cache_timeout=0):

        update_available = None
        if (show_hidden and engine.hidden) or not engine.hidden:
            if engine.installed:
                try:
                    if engine.up_to_date():
                        update_available = False
                    else:
                        update_available = True
                except BaseException as e:
                    click.secho(
                        f"[!]engine update failed for {engine.name}", fg="red", err=True
                    )
                    click.secho(f"[!]  {e}", fg="red", err=True)

            data.append(
                {
                    "name": engine.name,
                    "description": engine.description,
                    "installed": engine.installed,
                    "digest": engine.digest,
                    "global": True
                    if engine.name == engine_store.global_engine
                    else False,
                    "update_available": update_available,
                }
            )
    return data


@group.command(
    "describe-engine", short_help="", context_settings=dict(max_content_width=240)
)
@click.argument("name", required=False, type=click.STRING)
@exceptions.backend_handler()
@config.pass_options
def describe_engine(opts, name):
    """
    Provides a detailed description of an engine
    """
    if name:
        engine_name = name
    elif opts.engine:
        engine_name = opts.engine
    else:
        engine_name = engine_store.global_engine

    engine = engine_store.find_engine(engine_name)

    try:

        up_to_date = engine.up_to_date()
        latest_digest = engine.get_latest_digest()

    except BaseException as e:
        click.secho(
            f"[!] Engine update check failed for {engine.name}", fg="red", err=True
        )
        click.secho(f"[!]  {e}", fg="red", err=True)

        up_to_date = None
        latest_digest = None
        pass

    engine_details = {
        "engine": {
            "name": engine.name,
            "description": engine.description,
            "hidden": engine.hidden,
            "installed": engine.installed,
            "engine_dir": engine.engine_dir,
            "up_to_date": up_to_date,
            "current_digest": engine.digest,
            "latest_digest": latest_digest,
        },
        "environment": engine.environment,
        "install_state": engine.install_state,
    }

    sources = []
    for source in engine.sources:

        try:
            source_digest = source.digest
        except BaseException as e:
            click.secho(
                f"[!] Source check failed {engine.name} - {source.name}",
                fg="red",
                err=True,
            )
            click.secho(f"[!]  {e}", fg="red", err=True)

            source_digest = None
            pass

        sources.append(
            {
                "name": source.name,
                "description": source.description,
                "latest_digest": source_digest,
            }
        )

    parts = []
    for part in engine.parts:
        parts.append(
            {
                "type": part.type,
                "description": part.description,
                "source_path": part.source_path,
                "source_name": part.source_name,
            }
        )

    engine_details["sources"] = sources
    engine_details["parts"] = parts

    click.echo(json.dumps(engine_details, indent=2))


@group.command(
    "clean-engines", short_help="", context_settings=dict(max_content_width=240)
)
@click.argument("name", type=click.STRING, nargs=-1)
@exceptions.backend_handler()
def clean_engines(name):
    """
    Clean local engine store
    """
    if name:
        for engine in name:
            click.echo(f"[*] cleaning {engine} from {engine_store.store_dir}")
            engine_store.clean_engine(engine)

    else:
        click.echo(f"[*] cleaning all engines from {engine_store.store_dir}")
        engine_store.clean_engines()


@group.command(
    "install-engine", short_help="", context_settings=dict(max_content_width=240)
)
@click.option(
    "-u",
    "--update",
    is_flag=True,
    help="Update the engine",
)
@click.option(
    "-f",
    "--force",
    is_flag=True,
    help="Force reinstall of engine",
)
@click.argument("name", required=False, type=click.STRING)
@exceptions.backend_handler()
@pass_options
def install_engine(opts, name, force, update):
    """
    Install an engine

    Installs are designed to be idempotent so once installed, the -u
    or -f switches need to be used to install a more recent version
    of an existing engine.
    """

    if name is None:
        name = opts.engine or engine_store.global_engine

    try:
        engine = engine_store.find_engine(name, cache_timeout=0)

    except HamletEngineInvalidVersion as e:
        click.secho(
            (
                "[!] The state of this engine is not compatible with the cli\n"
                f"[!] Remove the engine using\n"
                f"     hamlet engine clean-engine {name}\n"
                "[!] then install the engine again"
            ),
            err=True,
            fg="red",
        )
        raise e

    if not engine.installed or force:
        click.echo(
            f"[*] installing engine | {name} | digest: {engine.get_latest_digest()}"
        )
        engine.install()
    elif not engine.up_to_date() and update:
        click.echo(f"[*] updating engine | {name}")
        click.echo(
            f"[*] current digest: {engine.digest} | latest digest: {engine.get_latest_digest()}"
        )
        engine.install()
    else:
        click.echo(f"[*] engine installed | {name} | digest: {engine.digest}")


@group.command(
    "set-engine", short_help="", context_settings=dict(max_content_width=240)
)
@click.argument("name", required=False, type=click.STRING)
@exceptions.backend_handler()
@config.pass_options
def set_engine(opts, name):
    """
    Sets the global engine used
    """
    name = name or opts.engine
    engine = engine_store.get_engine(name)

    click.echo(f"[*] global engine set to {engine.name}")
    engine_store.global_engine = engine.name


@group.command(
    "get-engine", short_help="", context_settings=dict(max_content_width=240)
)
@exceptions.backend_handler()
@config.pass_options
def get_engine(opts):
    """
    Gets the current global engine
    """
    name = opts.engine or engine_store.global_engine
    click.echo(engine_store.get_engine(name).name)


@group.command("env", short_help="", context_settings=dict(max_content_width=240))
@click.argument(
    "environment_variable",
    required=False,
    type=click.STRING,
)
@exceptions.backend_handler()
@config.pass_options
def env(opts, environment_variable):
    """
    Get the environment variables for the current engine
    """

    if opts.engine is None:
        engine = engine_store.get_engine(ENGINE_GLOBAL_NAME)
    else:
        engine = engine_store.get_engine(opts.engine)

    if environment_variable is None:
        click.echo("# run eval $(hamlet engine env) to set variables")
        for k, v in engine.environment.items():
            click.echo(f'export {k}="{v}"')

    else:
        try:
            click.echo(engine.environment[environment_variable])
        except KeyError:
            click.echo("")


@group.command(
    "add-engine-source-build",
    short_help="",
    context_settings=dict(max_content_width=240),
)
@click.option(
    "-p",
    "--path",
    type=click.Path(
        exists=True, file_okay=False, dir_okay=True, readable=True, resolve_path=True
    ),
    default=".",
    help="The path to generate build details for",
)
@exceptions.backend_handler()
def add_engine_source_build(path):
    """
    Generates build metadata for engine sources
    """

    build_details = EngineCodeSourceBuildData(path=path)

    hamlet_meta_dir = os.path.join(path, ".hamlet")
    hamlet_build_state_file = os.path.join(hamlet_meta_dir, "engine_source.json")
    if not os.path.isdir(hamlet_meta_dir):
        os.makedirs(hamlet_meta_dir)

    with open(hamlet_build_state_file, "w") as file:
        json.dump(build_details.details, file, indent=2)



import click
from tabulate import tabulate

from hamlet.command.common import exceptions
from hamlet.command.common.config import pass_options
from hamlet.command.common.display import json_or_table_option, wrap_text
from hamlet.backend.deploy import find_deployments


def deployments_table(data):
    tablerows = []
    for row in data:
        tablerows.append(
            [
                wrap_text(row["DeploymentGroup"]),
                wrap_text(row["DeploymentUnit"]),
                wrap_text(row["DeploymentProvider"]),
                wrap_text(row["CurrentState"]),
                wrap_text(row["District"]),
            ]
        )
    return tabulate(
        tablerows,
        headers=[
            "Deployment Group",
            "Deployment Unit",
            "Provider",
            "State",
            "District",
        ],
        tablefmt="github",
    )


@click.command(
    "list-deployments", short_help="", context_settings=dict(max_content_width=240)
)
@click.option(
    "-m",
    "--deployment-mode",
    default="update",
    help="The deployment mode to use for the deployment",
)
@click.option(
    "-l",
    "--deployment-group",
    default=".*",
    show_default=True,
    help="The deployment group pattern to match",
)
@click.option(
    "-u",
    "--deployment-unit",
    default=[".*"],
    show_default=True,
    multiple=True,
    help="The deployment unit pattern to match",
)
@click.option(
    "-s",
    "--deployment-state",
    type=click.Choice(
        ["deployed", "notdeployed", "orphaned"],
        case_sensitive=False,
    ),
    default=["deployed", "notdeployed", "orphaned"],
    multiple=True,
    help="The states of deployments to include",
)
@click.option(
    "-d",
    "--district",
    default=["segment"],
    multiple=True,
    help="The districts to include deployments from",
)
@json_or_table_option(deployments_table)
@exceptions.backend_handler()
@pass_options
def list_deployments(
    options,
    deployment_mode,
    deployment_group,
    deployment_unit,
    deployment_state,
    district,
):
    """
    List available deployments
    """
    return find_deployments(
        deployment_mode,
        deployment_group,
        deployment_units=deployment_unit,
        deployment_states=deployment_state,
        districts=district,
        **options.opts
    )



import os
from hamlet.backend import query as query_backend


class DescribeContext:
    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value


def query_occurrences_state(options, query, query_params=None, sub_query_text=None):
    query_args = {
        **options.opts,
        "generation_entrance": "occurrences",
        "output_filename": "occurrences-state.json",
        "use_cache": False,
    }
    query_result = query_backend.run(
        **query_args,
        cwd=os.getcwd(),
        query_text=query,
        query_params=query_params,
        sub_query_text=sub_query_text
    )

    return query_result



import os
import boto3
import json
import tempfile
from hamlet.backend.common.fsutils import Search
from .context_tree import find_gen3_root_dir, find_gen3_dirs, upgrade_cmdb, cleanup_cmdb
from .environment import Environment
from .utils import deep_dict_update


def set_context(cwd=None, environment_obj=None):
    e = environment_obj  # shortcut
    l = Environment()  # local env
    if e.GENERATION_CONTEXT_DEFINED:
        return
    e.GENERATION_CONTEXT_DEFINED = True
    l.GENERATION_CONTEXT_DEFINED_LOCAL = True
    if not e.GENERATION_TMPDIR:
        # *******************************************
        # * TODO: decide how to do temp dir cleanup *
        # *******************************************
        e.GENERATION_TMPDIR = tempfile.mkdtemp()
    e.GENERATION_DATA_DIR = find_gen3_root_dir(cwd)
    if not e.GENERATION_DATA_DIR:
        raise Exception("Can't locate the root of the directory tree.")
    if not e.GENERATION_NO_CMDB_CHECK:
        if not upgrade_cmdb(
            e.GENERATION_DATA_DIR, "", "", e.GENERATION_MAX_CMDB_UPGRADE_VERSION
        ):
            raise Exception("CMDB upgrade failed.")
        if not cleanup_cmdb(
            e.GENERATION_DATA_DIR, "", "", e.GENERATION_MAX_CMDB_UPGRADE_VERSION
        ):
            raise Exception("CMDB cleanup failed.")
    e.CACHE_DIR = os.path.join(e.GENERATION_DATA_DIR, "cache")
    os.makedirs(e.CACHE_DIR, exist_ok=True)

    l.TEMPLATE_COMPOSITES = ("account", "fragment")

    for composite in l.TEMPLATE_COMPOSITES:
        key = f"COMPOSITE_{composite.upper()}"
        e[key] = os.path.join(e.CACHE_DIR, f"{key.lower()}.ftl")

        if (
            not e.GENERATION_USE_CACHE
            and not e.GENERATION_USE_FRAGMENTS_CACHE
            or not Search.isfile(e.CACHE_DIR, "composite_account.ftl")
        ):
            l[f"{composite.lower()}_array"] = []
            # ********************************************
            # * TODO: add legacy start fragments section *
            # ********************************************
    # Check if the current directory gives any clue to the context
    # Accommodate both pre cmdb v2.0.0 where segment/environment in the config tree
    # and post v2.0.0 where they are in the infrastructure tree
    l.solutions_ancestor_dir = Search.upwards(cwd, "solutions")
    l.solutionsv2_ancestor_dir = Search.upwards(cwd, "solutionsv2")
    if not l.solutions_ancestor_dir and not l.solutionsv2_ancestor_dir:
        l.infrastructure_dir = cwd.replace("settings", "solutions")
        l.infrastructure_dir = l.infrastructure_dir.replace(
            "operations", "infrastructure"
        )
        l.infrastructure_dir = l.infrastructure_dir.replace("config", "infrastructure")
        if os.path.isdir(l.infrastructure_dir):
            cwd = l.infrastructure_dir
    if Search.isfile(cwd, "segment.json"):
        e.LOCATION = e.LOCATION or "segment"
        e.SEGMENT = Search.basename(cwd)
        if Search.isfile(cwd, "environment.json", up=1):
            cwd = Search.parent(cwd, up=1)
        else:
            e.ENVIRONMENT = e.SEGMENT
            e.SEGMENT = "default"
            cwd = os.path.join(Search.parent(cwd, up=3), "config")
    if Search.isfile(cwd, "environment.json"):
        e.LOCATION = e.LOCATION or "environment"
        e.ENVIRONMENT = Search.basename(cwd)
        cwd = os.path.join(Search.parent(cwd, up=3), "config")
    if Search.isfile(cwd, "account.json"):
        # account directory
        # We check it before checking for a product as the account directory
        # also acts as a product directory for shared infrastructure
        # An account directory may also have no product information e.g.
        # in the case of production environments in dedicated accounts.
        e.LOCATION = e.LOCATION or "account"
    if Search.isfile(cwd, "product.json"):
        if e.LOCATION == "account":
            e.LOCATION = "account|product"
        else:
            e.LOCATION = e.LOCATION or "product"
        e.PRODUCT = Search.basename(cwd)
        if e.PRODUCT == "config":
            e.PRODUCT = Search.basename(cwd, up=1)
    if Search.isfile(cwd, "integrator.json"):
        e.LOCATION = e.LOCATION or "integrator"
        e.INTEGRATOR = Search.basename(cwd)
    if (
        Search.isfile(cwd, "root.json")
        or Search.isdir(cwd, "config")
        and Search.isdir(cwd, "infrastructure")
    ):
        e.LOCATION = "root"
    cwd = e.GENERATION_DATA_DIR
    if not e.ACCOUNT:
        e.ACCOUNT = Search.basename(e.GENERATION_DATA_DIR)
    # Analyse directory structure
    find_gen3_dirs(e, e.GENERATION_DATA_DIR)
    # Build the composite solution ( aka blueprint)
    e.GENERATION_INPUT_SOURCE = e.GENERATION_INPUT_SOURCE or "composite"
    if e.GENERATION_INPUT_SOURCE == "composite":
        l.blueprint_alternate_dirs = (
            e.SEGMENT_SOLUTIONS_DIR,
            e.ENVIRONMENT_SHARED_SOLUTIONS_DIR,
            e.SEGMENT_SHARED_SOLUTIONS_DIR,
            e.PRODUCT_SHARED_SOLUTIONS_DIR,
        )
        e.COMPOSITE_BLUEPRINT = os.path.join(e.CACHE_DIR, "composite_blueprint.json")
        if (
            not e.GENERATION_USE_CACHE
            and not e.GENERATION_USE_BLUEPRINT_CACHE
            or not os.path.isfile(e.COMPOSITE_BLUEPRINT)
        ):
            l.blueprint_array = []
            for blueprint_alternate_dir in l.blueprint_alternate_dirs:
                if not blueprint_alternate_dir or not os.path.isdir(
                    blueprint_alternate_dir
                ):
                    continue
                l.blueprint_array += Search.match_files(
                    os.path.join("**", "segment*.json"),
                    os.path.join("**", "environment*.json"),
                    os.path.join("**", "solution*.json"),
                    os.path.join("**", "domains*.json"),
                    os.path.join("**", "ipaddressgroups*.json"),
                    os.path.join("**", "countrygroups*.json"),
                    root=blueprint_alternate_dir,
                )
            l.blueprint_array += Search.match_files(
                os.path.join("**", "domains*.json"),
                os.path.join("**", "ipaddressgroups*.json"),
                os.path.join("**", "countrygroups*.json"),
                os.path.join("**", "profiles*.json"),
                os.path.join("**", "product*.json"),
                root=e.PRODUCT_DIR,
            )
            l.blueprint_array += Search.match_files(
                os.path.join("**", "domains*.json"),
                os.path.join("**", "ipaddressgroups*.json"),
                os.path.join("**", "countrygroups*.json"),
                os.path.join("**", "account*.json"),
                root=e.ACCOUNT_DIR,
            )
            l.blueprint_array += Search.match_files(
                os.path.join("**", "domains*.json"),
                os.path.join("**", "ipaddressgroups*.json"),
                os.path.join("**", "countrygroups*.json"),
                os.path.join("**", "profiles*.json"),
                os.path.join("**", "tenant*.json"),
                root=e.TENANT_DIR,
            )
            if not l.blueprint_array:
                blueprint = {}
                with open(e.COMPOSITE_BLUEPRINT, "wt+") as f:
                    f.write("{}")
            else:
                #  merging blueprint components
                blueprint = {}
                for blueprint_component_json in l.blueprint_array:
                    with open(blueprint_component_json, "rt") as f:
                        deep_dict_update(blueprint, json.load(f))
                with open(e.COMPOSITE_BLUEPRINT, "wt+") as f:
                    json.dump(blueprint, f, indent=4, sort_keys=True)

            # Extract key settings from the composite solution
            tenant = blueprint.get("Tenant", {})
            account = blueprint.get("Account", {})
            product = blueprint.get("Product", {})
            segment = blueprint.get("Segment", {})
            e.TID = tenant.get("Id")
            e.TENANT = tenant.get("Name")
            e.AID = account.get("Id")
            e.AWSID = account.get("AWSId")
            e.ACCOUNT_REGION = account.get("Region")
            e.PID = product.get("Id")
            e.PRODUCT_REGION = product.get("Region")
            e.DEPLOYMENTUNIT_REGION = product.get(e.DEPLOYMENT_UNIT, {}).get("Region")
            e.SID = segment.get("Id")

            e.COMPONENT_REGION = e.DEPLOYMENTUNIT_REGION or e.PRODUCT_REGION
            e.REGION = e.REGION or e.COMPONENT_REGION
            # Perform a few consistency checks
            if not e.REGION:
                raise Exception(
                    "The region must be defined in the Product blueprint section."
                )

            l.BLUEPRINT_ACCOUNT = account.get("Name") or account.get("Id")
            l.BLUEPRINT_PRODUCT = product.get("Name") or product.get("Id")
            l.BLUEPRINT_SEGMENT = segment.get("Name") or segment.get("Id")
            if (
                e.ACCOUNT
                and l.BLUEPRINT_ACCOUNT != "Account"
                and e.ACCOUNT != l.BLUEPRINT_ACCOUNT
            ):
                raise Exception(
                    f"Blueprint account of {l.BLUEPRINT_ACCOUNT} doesn't match expected value of {e.ACCOUNT}"
                )
            if (
                e.PRODUCT
                and l.BLUEPRINT_PRODUCT != "Product"
                and e.PRODUCT != l.BLUEPRINT_PRODUCT
            ):
                raise Exception(
                    f"Blueprint product of {l.BLUEPRINT_PRODUCT} doesn't match expected value of ${e.PRODUCT}"
                )
            if (
                e.SEGMENT
                and l.BLUEPRINT_SEGMENT != "Segment"
                and e.SEGMENT != l.BLUEPRINT_SEGMENT
            ):
                raise Exception(
                    f"Blueprint segment of {e.BLUEPRINT_SEGMENT} doesn't match expected value of {e.SEGMENT}"
                )
    # Set default AWS credentials if available (hook from Jenkins framework)
    l.CHECK_AWS_ACCESS_KEY_ID = (
        e.AWS_ACCESS_KEY_ID
        or e.ACCOUNT_TEMP_AWS_ACCESS_KEY_ID
        or e[e.ACCOUNT_AWS_ACCESS_KEY_ID_VAR]
    )
    if l.CHECK_AWS_ACCESS_KEY_ID:
        e.AWS_ACCESS_KEY_ID = l.CHECK_AWS_ACCESS_KEY_ID

    l.CHECK_AWS_SECRET_ACCESS_KEY = (
        e.AWS_SECRET_ACCESS_KEY
        or e.ACCOUNT_TEMP_AWS_SECRET_ACCESS_KEY
        or e[e.ACCOUNT_AWS_SECRET_ACCESS_KEY_VAR]
    )
    if l.CHECK_AWS_SECRET_ACCESS_KEY:
        e.AWS_SECRET_ACCESS_KEY = l.CHECK_AWS_SECRET_ACCESS_KEY

    l.CHECK_AWS_SESSION_TOKEN = e.AWS_SESSION_TOKEN or e.ACCOUNT_TEMP_AWS_SESSION_TOKEN
    if l.CHECK_AWS_SESSION_TOKEN:
        e.AWS_SESSION_TOKEN = l.CHECK_AWS_SESSION_TOKEN

    # Set the profile for IAM access if AWS credentials not in the environment
    if not e.AWS_ACCESS_KEY_ID or not e.AWS_SECRET_ACCESS_KEY:
        available_profiles = boto3.session.Session().available_profiles
        if e.ACCOUNT and e.ACCOUNT in available_profiles:
            e.AWS_DEFAULT_PROFILE = e.ACCOUNT
        if e.AID and e.AID in available_profiles:
            e.AWS_DEFAULT_PROFILE = e.AID
        if e.AWSID and e.AWSID in available_profiles:
            e.AWS_DEFAULT_PROFILE = e.AWSID



import click
import functools
import os


def common_generate_options(f):
    """Add common options for the generate command group"""

    @functools.wraps(f)
    @click.option(
        "--output-dir",
        "-o",
        help="The directory to save the generated content",
        show_default=True,
        type=click.Path(file_okay=False, dir_okay=True, writable=True),
        default=os.getcwd(),
    )
    def wrapper(*args, **kwargs):
        f(*args, **kwargs)

    return wrapper



import click
import json
import jmespath
from jmespath.exceptions import JMESPathError

from hamlet.command.common import config, exceptions
from hamlet.command.common.display import json_or_table_option, table_from_dict

from .common import query_occurrences_state, DescribeContext


DESCRIBE_OCCURRENCE_QUERY = "Occurrences[?Core.TypedRawName=={name}] | [0]"


def query_occurrence_state(ctx, query=None):
    """
    Query the state of a single occurrence
    """
    options_context = ctx.ensure_object(config.Options)
    describe_context = ctx.ensure_object(DescribeContext)

    base_result = query_occurrences_state(
        options=options_context,
        query=DESCRIBE_OCCURRENCE_QUERY,
        query_params={"name": describe_context.name},
    )

    if base_result is None:
        raise exceptions.BackendException(
            f"Could not find occurrence {describe_context.name}"
        )

    if query:
        try:
            return jmespath.search(query, base_result)
        except JMESPathError as e:
            raise exceptions.BackendException(f"JMESPath query error: {str(e)}") from e
    else:
        return base_result


@click.group(
    "describe-occurrence",
    short_help="",
    context_settings=dict(max_content_width=240),
    invoke_without_command=True,
)
@click.option(
    "-n", "--name", required=True, help="The name of the occurrence to describe"
)
@click.option(
    "-q",
    "--query",
    help="A JMESPath query to apply to the results",
)
@exceptions.backend_handler()
@click.pass_context
def describe_occurrence(ctx, name, query):
    """
    Describe an occurrence
    """

    describe_context = DescribeContext()
    describe_context.name = name
    ctx.obj = describe_context

    if ctx.invoked_subcommand is None:
        click.echo(json.dumps(query_occurrence_state(ctx, query), indent=4))


@describe_occurrence.command(
    "solution", short_help="", context_settings=dict(max_content_width=240)
)
@exceptions.backend_handler()
@click.pass_context
def describe_occurrence_solution(ctx):
    """
    Predefined query:
    The evaluated solution configuration for the occurrence
    """
    result = query_occurrence_state(ctx, "Configuration.Solution")
    click.echo(json.dumps(result, indent=4))


@describe_occurrence.command(
    "resources", short_help="", context_settings=dict(max_content_width=240)
)
@exceptions.backend_handler()
@click.pass_context
def describe_occurrence_resources(ctx):
    """
    Predefined query:
    The resources of the occurrence
    """
    result = query_occurrence_state(ctx, "State.Resources")
    click.echo(json.dumps(result, indent=4))


@describe_occurrence.command(
    "setting-namespaces", short_help="", context_settings=dict(max_content_width=240)
)
@exceptions.backend_handler()
@click.pass_context
def describe_occurrence_setting_namespaces(ctx):
    """
    Predefined query:
    The setting namespaces for the occurrence
    """
    result = query_occurrence_state(ctx, "Configuration.SettingNamespaces")
    click.echo(json.dumps(result, indent=4))


@describe_occurrence.command(
    "attributes", short_help="", context_settings=dict(max_content_width=240)
)
@json_or_table_option(table_from_dict)
@exceptions.backend_handler()
@click.pass_context
def describe_occurrence_attributes(ctx):
    """
    Predefined query:
    The attributes for the occurrence
    """
    result = query_occurrence_state(ctx, "State.Attributes")
    return result



import collections
from unittest import mock
from click.testing import CliRunner
from hamlet.command.generate.cmdb import generate_product
from tests.unit.command.generate.test_generate_command import run_generate_command_test

OPTIONS = collections.OrderedDict()
OPTIONS["!--product-id,product_id"] = "test-product-id"
OPTIONS["--product-name,product_name"] = ("test-product-name", "test-product-id")
OPTIONS["--dns-zone,dns_zone"] = ("test-zone-name", "")
OPTIONS["--environment-id,environment_id"] = (
    "test-environment-id",
    "test-environment-id",
)
OPTIONS["--environment-name,environment_name"] = (
    "test-environment-name",
    "test-environment-name",
)
OPTIONS["--segment-id,segment_id"] = ("test-segment-id", "default")
OPTIONS["--segment-name,segment_name"] = ("test-segment-name", "default")


@mock.patch("hamlet.command.generate.cmdb.generate_product_backend")
def test(generate_prodcut_backend):
    run_generate_command_test(
        CliRunner(), generate_product, generate_prodcut_backend.run, OPTIONS
    )



import click
import typing
import os

from click_configfile import ConfigFileReader, SectionSchema, matches_section
from hamlet.env import HAMLET_GLOBAL_CONFIG

from hamlet.utils import ConfigParam

from hamlet.backend.engine.engine_loader import EngineLoader

from hamlet.backend.engine.engine import Engine
from hamlet.backend.engine.engine_source import (
    LocalDirectoryEngineSource,
    ContainerEngineSource,
)
from hamlet.backend.engine.engine_part import (
    CoreEnginePart,
    AWSEnginePluginPart,
    AzureEnginePluginPart,
    CMDBEnginePluginPart,
    BashExecutorEnginePart,
    WrapperEnginePart,
)


def get_engine_config_dir():
    print("getting engine config")
    return HAMLET_GLOBAL_CONFIG.config_dir


class EngineSchema(object):
    """Defines the schema for engines loaded through a toml based config file"""

    @matches_section("engine:*")
    class Engine(SectionSchema):
        """The engine is the collection of sources and parts that provide an overall hamlet install"""

        #: A general description of the engine
        description = ConfigParam(type=str)

        #: A list of source names defined in engine_source: sections
        sources = ConfigParam(type=str, multiple=True)

        #: A list of part names defined in engine_part: sections
        parts = ConfigParam(type=str, multiple=True)

    @matches_section("engine_source:*")
    class EngineSource(SectionSchema):
        """Engine source configuration"""

        #: A general description of the engine_source
        description = ConfigParam(type=str)

        #: The type of engine_source defines where to get the source from
        type = ConfigParam(type=click.Choice(["local_dir", "container"]))

        #: local_dir type: the local path to the source
        local_dir_path = ConfigParam(
            type=click.Path(dir_okay=True, file_okay=False, readable=True)
        )

        #: container type: the base url for the container registry
        container_registry_url = ConfigParam(type=str)

        #: container type: the name of the container repository
        container_repository = ConfigParam(type=str)

        #: container type: the tag of the image to pull from the repository
        container_tag = ConfigParam(type=str)

    @matches_section("engine_part:*")
    class EnginePart(SectionSchema):
        """Engine part configuration"""

        #: The type of the engine_part defines what function a particular part of the source does in hamlet
        type = ConfigParam(
            type=click.Choice(
                [
                    "engine",
                    "engine-plugin-aws",
                    "engine-plugin-azure",
                    "engine-plugin-cmdb",
                    "executor-bash",
                    "wrapper",
                ]
            )
        )

        #: The name of the source in engine_source: sections that the part belongs to
        source_name = ConfigParam(type=str)

        #: The source_path is the path within the source where the part functions are located
        source_path = ConfigParam(type=str)


class EngineReader(ConfigFileReader):

    """Reader for local engine configuration."""

    config_files = ["engine.ini", "engine"]
    config_name = "engine"
    config_searchpath = []
    config_section_schemas = [
        EngineSchema.Engine,
        EngineSchema.EnginePart,
        EngineSchema.EngineSource,
    ]

    @classmethod
    def select_config_schema_for(cls, section_name):
        section_schema = super(EngineReader, cls).select_config_schema_for(section_name)
        for v in section_schema.__dict__:
            if isinstance(v, ConfigParam):
                v.ctx = cls
        return section_schema

    @classmethod
    def load_engines(cls):
        """Load engines found in the config file."""

        if os.path.exists(HAMLET_GLOBAL_CONFIG.config_dir):
            if os.path.isdir(HAMLET_GLOBAL_CONFIG.config_dir):
                cls.config_searchpath.insert(0, HAMLET_GLOBAL_CONFIG.config_dir)

        config = cls.read_config()

        for k, v in config.items():
            if k.startswith("engine:"):
                engine = Engine(
                    name=k[len("engine:") :],
                    description=v.get("description", None),
                )

                sources = []
                for source_name in v.get("sources"):
                    try:
                        source_config = config[f"engine_source:{source_name}"]

                    except KeyError as e:
                        e.msg = (
                            f"The provided engine source {source_name}"
                            "could not be found in the configured local sources"
                        )
                        raise e

                    if source_config["type"] == "local_dir":
                        sources.append(
                            LocalDirectoryEngineSource(
                                name=source_name,
                                description=source_config.get("description", None),
                                env_path=source_config["local_dir_path"],
                            )
                        )

                    if source_config["type"] == "container":
                        sources.append(
                            ContainerEngineSource(
                                name=source_name,
                                description=source_config.get("description", None),
                                registry_url=source_config.get(
                                    "container_registry_url"
                                ),
                                repository=source_config.get("container_repository"),
                                tag=source_config.get("container_tag"),
                            )
                        )

                parts = []
                for part_name in v.get("parts"):

                    part_config = config[f"engine_part:{part_name}"]

                    part_source_config = {
                        "source_name": part_config["source_name"],
                        "source_path": part_config.get("source_path", ""),
                    }

                    if part_config["type"] == "engine":
                        parts.append(CoreEnginePart(**part_source_config))

                    if part_config["type"] == "engine-plugin-aws":
                        parts.append(AWSEnginePluginPart(**part_source_config))

                    if part_config["type"] == "engine-plugin-azure":
                        parts.append(AzureEnginePluginPart(**part_source_config))

                    if part_config["type"] == "engine-plugin-cmdb":
                        parts.append(CMDBEnginePluginPart(**part_source_config))

                    if part_config["type"] == "executor-bash":
                        parts.append(BashExecutorEnginePart(**part_source_config))

                    if part_config["type"] == "wrapper":
                        parts.append(WrapperEnginePart(**part_source_config))

                engine.sources = sources
                engine.parts = parts

                yield engine


class UserDefinedEngineLoader(EngineLoader):
    """
    Loads a user defined configuration file defining engines and their configuration
    """

    def __init__(self):
        super().__init__()

    def load(self) -> typing.Iterable[Engine]:

        engine_cls = EngineReader
        for engine in engine_cls.load_engines():
            yield engine



import os
import shutil
import json
from datetime import datetime

from .common import ENGINE_STORE_DEFAULT_DIR, ENGINE_GLOBAL_NAME
from .exceptions import EngineStoreMissingEngineException

from .loaders.base import GlobalEngineLoader, InstalledEngineLoader
from .loaders.unicycle import UnicycleEngineLoader
from .loaders.tram import LatestTramEngineLoader, TramEngineLoader
from .loaders.train import LatestTrainEngineLoader, TrainEngineLoader
from .loaders.user import UserDefinedEngineLoader


class EngineStore:
    """
    The EngineStore manages the collection of engines that can be used for hamlet
    Loaders are assigned to the engine store which provide engines to the store
    It manages the location of the engine installations and maintains a persistant state
    that tracks global properties that effect all engines
    """

    def __init__(self, store_dir):
        self._engines = {}
        self.store_dir = store_dir
        self._store_state_file = os.path.join(self.store_dir, "store_state.json")
        self.engine_dir = os.path.join(self.store_dir, "engines")

        self.store_state = {}
        self.load_store_state()

        self._global_engine = self.store_state.get("global_engine", None)

        self.local_engine_loaders = [
            InstalledEngineLoader(engine_dir=self.engine_dir),
            UserDefinedEngineLoader(),
            GlobalEngineLoader(),
        ]

        self.external_engine_loaders = [
            UnicycleEngineLoader(),
            LatestTramEngineLoader(),
            TramEngineLoader(),
            LatestTrainEngineLoader(),
            TrainEngineLoader(),
        ]

    @property
    def global_engine(self):
        """
        The global engine defines a special engine that uses other engines to provide parts
        the global provides a consistent location that will link to other engine installations
        """
        return self._global_engine

    @global_engine.setter
    def global_engine(self, name):
        """
        Setting the global engine will locate the provided engine then update
        its symlinks to point to the provided engine.
        """
        global_engine = self.get_engine(ENGINE_GLOBAL_NAME)
        engine = self.get_engine(name)

        for type, path in global_engine.part_paths.items():
            if os.path.islink(path):
                os.unlink(path)

            if os.path.isdir(path):
                os.rmdir(path)

            if engine.part_paths.get(type, None):
                os.symlink(engine.part_paths[type], path, target_is_directory=True)
            else:
                os.makedirs(path)

        self._global_engine = engine.name

        self.store_state = {"global_engine": self.global_engine}
        self.save_store_state()

    def _load_local_engines(self):
        for loader in self.local_engine_loaders:
            for engine in loader.load():
                engine.engine_dir = self.engine_dir
                self._engines[engine.name] = engine

    def _load_external_engines(self, cache_timeout):
        if (
            datetime.now()
            - datetime.strptime(
                self.store_state.get(
                    "last_external_load", datetime.now().isoformat(timespec="seconds")
                ),
                "%Y-%m-%dT%H:%M:%S",
            )
        ).seconds >= cache_timeout:

            for loader in self.external_engine_loaders:
                for engine in loader.load():
                    engine.engine_dir = self.engine_dir
                    self._engines[engine.name] = engine

            self.store_state["last_external_load"] = datetime.now().isoformat(
                timespec="seconds"
            )
            self.save_store_state()

    def get_engines(self):
        """
        Return as list of the engines available locally
        """
        self._load_local_engines()
        return list(self._engines.values())

    def get_engine(self, name, allow_missing=False):
        """
        Return an existing engine
        """
        self._load_local_engines()
        result = self._engines.get(name, None)

        if not allow_missing and not result:
            raise EngineStoreMissingEngineException(
                (
                    f"[!] Could not find engine {name} in local engine store\n"
                    "[!] Check that the engine is installed and the name is correct\n"
                    f"[!] Run hamlet engine install-engine {name} to install it\n"
                )
            )

        return result

    def find_engines(self, cache_timeout=0):
        """
        Find engines that can be installed from external sources
        """
        self._load_external_engines(cache_timeout)
        return list(self._engines.values())

    def find_engine(self, name, allow_missing=False, cache_timeout=0):
        """
        Find an external engine
        """
        self._load_external_engines(cache_timeout)
        result = self._engines.get(name, None)

        if not allow_missing and not result:
            raise EngineStoreMissingEngineException(
                (
                    f"[!] Could not find an engine matching engine the name {name}\n"
                    "[!] Check that the name of the engine is right\n"
                    "[!] Run hamlet engine list-engines for available engines\n"
                )
            )

        return result

    def load_store_state(self):
        """
        Load the persisted enginestore state
        """
        if os.path.isfile(self._store_state_file):
            with open(self._store_state_file, "r") as file:
                self.store_state = json.load(file)

    def save_store_state(self):
        """
        Save the persisted enginestore state
        """
        if self.store_state:
            if not os.path.isdir(self.store_dir):
                os.makedirs(self.store_dir)

            with open(self._store_state_file, "w") as file:
                json.dump(self.store_state, file)

    def clean_engines(self):
        if os.path.isdir(os.path.join(self.engine_dir)):
            shutil.rmtree(self.engine_dir)

    def clean_engine(self, name):
        if os.path.isdir(os.path.join(self.engine_dir, name)):
            shutil.rmtree(os.path.join(self.engine_dir, name))


engine_store = EngineStore(store_dir=ENGINE_STORE_DEFAULT_DIR)



# NOTE: this file must remain valid python file in order to perform tests on it


def cfn_lint_test(filename, ignore_checks=None):
    import json
    import subprocess

    cmd_args = ["cfn-lint", "-f", "json", filename]

    if ignore_checks is not None:
        cmd_args.append("--ignore-checks")
        cmd_args.append(" ".join(ignore_checks))

    cmd = " ".join(cmd_args)

    result = subprocess.run(
        cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, encoding="utf8"
    )
    if result.stderr:
        raise Exception(result.stderr)
    else:
        errors = json.loads(result.stdout)
        if errors:
            raise AssertionError(json.dumps(errors, indent=4))



import os
from jinja2 import Environment, FileSystemLoader

TEMPLATES_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "templates")


loader = Environment(
    loader=FileSystemLoader(TEMPLATES_DIR), trim_blocks=True, lstrip_blocks=True
)



import os
import json
import tempfile
from unittest import mock

# import subprocess
from hamlet.backend.generate.cmdb import account, tenant, product
from hamlet.backend.create.template.environment import Environment
from hamlet.backend.create.template import context_tree as ct
from .conftest import conf


def test_find_gen3_dirs(clear_cmdb, cmdb):
    clear_cmdb()

    tenant_name = conf["cmdb"]["tenant"]["tenant_name"]
    account_name = conf["cmdb"]["account"]["account_name"]
    product_name = conf["cmdb"]["product"]["product_name"]
    segment_name = conf["cmdb"]["product"]["segment_name"]
    environment_name = conf["cmdb"]["product"]["environment_name"]

    with mock.patch.dict(os.environ, {"ROOT_DIR": cmdb.ROOT_DIR}):

        with cmdb() as path:
            tenant.run(**conf["cmdb"]["tenant"], output_dir=path())
        with cmdb("accounts") as path:
            account.run(**conf["cmdb"]["account"], output_dir=path())
        with cmdb() as path:
            product.run(**conf["cmdb"]["product"], output_dir=path())

        with cmdb() as path:
            root_dir = path()
            assert root_dir == ct.find_gen3_root_dir(path())
            tenant_dir = os.path.join(root_dir, "accounts", tenant_name)
            assert tenant_dir == ct.find_gen3_tenant_dir(root_dir, tenant_name)
            account_base_dir = os.path.join(root_dir, "accounts", account_name)
            account_dir = os.path.join(account_base_dir, "config")
            account_state_dir = os.path.join(account_base_dir, "infrastructure")
            assert account_dir == ct.find_gen3_account_dir(root_dir, account_name)
            assert account_state_dir == ct.find_gen3_account_state_dir(
                root_dir, account_name
            )
            product_base_dir = os.path.join(root_dir, product_name)
            product_dir = os.path.join(product_base_dir, "config")
            product_infr_dir = os.path.join(product_base_dir, "infrastructure")
            environment_dir = os.path.join(product_dir, "solutionsv2", environment_name)
            assert product_dir == ct.find_gen3_product_dir(root_dir, product_name)
            assert product_infr_dir == ct.find_gen3_product_infrastructure_dir(
                root_dir, product_name
            )
            assert environment_dir == ct.find_gen3_environment_dir(
                root_dir, product_name, environment_name
            )
            e = Environment()
            ct.find_gen3_dirs(
                e,
                root_dir,
                tenant=tenant_name,
                account=account_name,
                product=product_name,
                environment=environment_name,
                segment=segment_name,
            )
            assert e.ROOT_DIR == root_dir
            assert e.TENANT_DIR == tenant_dir

            assert e.ACCOUNT_DIR == account_dir
            assert e.ACCOUNT_STATE_DIR == account_state_dir
            assert e.ACCOUNT_SETTINGS_DIR == os.path.join(account_dir, "settings")
            assert e.ACCOUNT_OPERATIONS_DIR == os.path.join(
                account_state_dir, "operations"
            )
            product_settings_dir = os.path.join(product_dir, "settings")
            product_operations_dir = os.path.join(product_infr_dir, "operations")
            product_solutions_dir = os.path.join(product_dir, "solutionsv2")
            environment_shared_operations_dir = os.path.join(
                product_infr_dir, "operations", environment_name
            )
            environment_shared_settings_dir = os.path.join(
                product_dir, "settings", environment_name
            )
            environment_shared_solutions_dir = os.path.join(
                product_dir, "solutionsv2", environment_name
            )
            segment_solutions_dir = os.path.join(
                product_dir, "solutionsv2", environment_name, segment_name
            )
            segment_settings_dir = os.path.join(
                product_dir, "settings", environment_name, segment_name
            )
            segment_operations_dir = os.path.join(
                product_infr_dir, "operations", environment_name, segment_name
            )
            segment_builds_dir = segment_settings_dir
            segment_shared_settings_dir = os.path.join(
                product_dir, "settings", "shared", segment_name
            )
            segment_shared_solutions_dir = os.path.join(
                product_dir, "solutionsv2", "shared", segment_name
            )
            segment_shared_operations_dir = os.path.join(
                product_infr_dir, "operations", "shared", segment_name
            )
            assert e.PRODUCT_DIR == product_dir
            assert e.PRODUCT_SETTINGS_DIR == product_settings_dir
            assert e.PRODUCT_INFRASTRUCTURE_DIR == product_infr_dir
            assert e.PRODUCT_OPERATIONS_DIR == product_operations_dir
            assert e.PRODUCT_SOLUTIONS_DIR == product_solutions_dir
            assert e.PRODUCT_SHARED_SETTINGS_DIR == os.path.join(
                product_settings_dir, "shared"
            )
            assert e.PRODUCT_SHARED_OPERATIONS_DIR == os.path.join(
                product_operations_dir, "shared"
            )
            assert e.PRODUCT_SHARED_SOLUTIONS_DIR == os.path.join(
                product_solutions_dir, "shared"
            )
            assert e.ENVIRONMENT_SHARED_SETTINGS_DIR == environment_shared_settings_dir
            assert (
                e.ENVIRONMENT_SHARED_SOLUTIONS_DIR == environment_shared_solutions_dir
            )
            assert (
                e.ENVIRONMENT_SHARED_OPERATIONS_DIR == environment_shared_operations_dir
            )
            assert e.SEGMENT_SOLUTIONS_DIR == segment_solutions_dir
            assert e.SEGMENT_SETTINGS_DIR == segment_settings_dir
            assert e.SEGMENT_OPERATIONS_DIR == segment_operations_dir
            assert e.SEGMENT_BUILDS_DIR == segment_builds_dir
            assert e.SEGMENT_SHARED_SETTINGS_DIR == segment_shared_settings_dir
            assert e.SEGMENT_SHARED_SOLUTIONS_DIR == segment_shared_solutions_dir
            assert e.SEGMENT_SHARED_OPERATIONS_DIR == segment_shared_operations_dir
            print(e)


class FSNode:
    def __init__(self, path):
        self.path = path

    def __getitem__(self, key):
        return self.__class__(os.path.join(self.path, key))

    def __setitem__(self, key, value):
        if not isinstance(value, str):
            value = json.dumps(value, indent=4)
        os.makedirs(self.path, exist_ok=True)
        with open(os.path.join(self.path, key), "wt+") as f:
            f.write(value)

    def exists(self):
        return os.path.exists(self.path)

    def mkdir(self):
        os.makedirs(self.path, exist_ok=True)
        return self

    def touchfile(self):
        os.makedirs(os.path.dirname(self.path), exist_ok=True)
        open(self.path, "w").close()
        return self

    def mkfile(self, data):
        os.makedirs(os.path.dirname(self.path), exist_ok=True)
        with open(self.path, "wt+") as f:
            f.write(data)
        return self

    def mkjson(self, data):
        data = json.dumps(data)
        self.mkfile(data)
        return self

    def isfile(self):
        return os.path.isfile(self.path)

    def isdir(self):
        return os.path.isdir(self.path)

    def json(self):
        if self.isfile():
            with open(self.path, "rt") as f:
                return json.load(f)
        raise TypeError(f"{self.path} is not a file")

    def text(self):
        if self.isfile():
            with open(self.path, "rt") as f:
                return f.read()
        raise TypeError(f"{self.path} is not a file")


def test_upgrade_version_1_0_0():
    def create_test_fs(root):
        root = FSNode(root)

        root["build.ref"] = "commit1 tag1"
        root["builds"]["build.ref"] = "commit2 tag2"

        root["shared-build.ref"] = "reference1"
        root["shared-builds"]["shared-build.ref"] = "reference2"

        root["credentials.json"] = {"Credentials": ["1", "2"]}
        root["credentials"]["credentials.json"] = {"Credentials": ["3", "4"]}

        root["container.json"] = {"Id": "1"}
        root["containers"]["container.json"] = {"Id": "2"}

    def assert_post_upgrade_structure(root):
        root = FSNode(root)
        # Legacy
        assert not root["build.ref"].exists()
        assert not root["builds"]["build.ref"].exists()

        assert not root["shared-build.ref"].exists()
        assert not root["shared-builds"]["shared-build.ref"].exists()

        assert not root["container.json"].exists()
        assert not root["containers"]["container.json"].exists()
        # Upgraded
        assert root["build.json"].json() == {
            "Commit": "commit1",
            "Tag": "tag1",
            "Formats": ["docker"],
        }
        assert root["builds"]["build.json"].json() == {
            "Commit": "commit2",
            "Tag": "tag2",
            "Formats": ["docker"],
        }

        assert root["shared_build.json"].json() == {"Reference": "reference1"}
        assert root["shared-builds"]["shared_build.json"].json() == {
            "Reference": "reference2"
        }

        assert root["credentials.json"].json() == ["1", "2"]
        assert root["credentials"]["credentials.json"].json() == ["3", "4"]

        assert root["segment.json"].json() == {"Id": "1"}
        assert root["containers"]["segment.json"].json() == {"Id": "2"}

    with tempfile.TemporaryDirectory() as root:
        create_test_fs(root)
        assert ct.upgrade_cmdb_repo_to_v1_0_0(root, "")
        assert ct.cleanup_cmdb_repo_to_v1_0_0(root, "")
        assert_post_upgrade_structure(root)


def test_upgrade_version_1_1_0():
    def create_test_fs(root):
        root = FSNode(root)

        appsettings = root["appsettings"]
        appsettings["appsettings-file.json"].touchfile()
        appsettings["appsettings-env"]["subdir"][
            "appsettings-env-sub-file.json"
        ].touchfile()
        appsettings["appsettings-env"]["appsettings-env-file.json"].touchfile()

        solutions = root["solutions"]
        solutions["solutions-file.json"].touchfile()
        solutions["segment.json"] = {
            "Segment": {
                "Environment": "solutions",
                "Id": "Id",
                "Name": "Name",
                "Title": "Title",
                "NonLegacyKey": "NonLegacyKey",
            }
        }
        solutions["solutions-env"]["subdir"]["solutions-env-sub-file.json"].touchfile()
        solutions["solutions-env"]["solutions-env-file.json"].touchfile()
        solutions["solutions-env"]["segment.json"] = {
            "Segment": {
                "Environment": "solutions-env",
                "Id": "Id",
                "Name": "Name",
                "Title": "Title",
                "NonLegacyKey": "NonLegacyKey",
            }
        }

        credentials = root["credentials"]
        credentials["credentials-file.json"].touchfile()
        credentials["aws-ssh.pem"].touchfile()
        credentials["credentials-env"]["subdir"][
            "credentials-env-sub-file.json"
        ].touchfile()
        credentials["credentials-env"]["credentials-env-file.json"].touchfile()
        credentials["credentials-env"][".gitignore"] = "test"
        credentials["credentials-env"]["aws-ssh-env.pem"].touchfile()

        aws = root["aws"]
        aws["aws-file.json"].touchfile()
        aws["cf"]["aws-cf-file.json"].touchfile()
        aws["aws-env"]["subdir"]["aws-env-sub-file.json"].touchfile()
        aws["aws-env"]["cf"]["aws-env-cf-file.json"].touchfile()
        aws["aws-env"]["cf"]["subdir"]["aws-env-cf-sub-file.json"].touchfile()

    def assert_post_upgrade_structure(root):
        root = FSNode(root)
        # Legacy
        assert not root["appsettings"].exists()
        # Upgraded
        settings = root["settings"]
        assert settings["shared"]["appsettings-file.json"].isfile()
        assert settings["appsettings-env"]["default"]["subdir"][
            "appsettings-env-sub-file.json"
        ].isfile()
        assert settings["appsettings-env"]["default"][
            "appsettings-env-file.json"
        ].isfile()
        # Legacy
        assert not root["solutions"].exists()
        # Upgraded
        solutionsv2 = root["solutionsv2"]
        assert solutionsv2.isdir()
        assert solutionsv2["shared"]["default"]["solutions-file.json"].isfile()
        assert solutionsv2["shared"]["default"]["segment.json"].json() == {
            "Segment": {"Id": "default"}
        }
        assert solutionsv2["solutions-env"]["default"]["subdir"][
            "solutions-env-sub-file.json"
        ].isfile()
        assert solutionsv2["solutions-env"]["default"][
            "solutions-env-file.json"
        ].isfile()
        assert solutionsv2["solutions-env"]["default"]["segment.json"].json() == {
            "Segment": {"NonLegacyKey": "NonLegacyKey"}
        }
        assert solutionsv2["solutions-env"]["environment.json"].json() == {
            "Environment": {"Id": "solutions-env"}
        }
        # Legacy
        assert not root["credentials"].exists()
        # Upgraded
        operations = root["operations"]
        assert operations["shared"][".aws-ssh.pem"].isfile()
        assert operations["shared"][".gitignore"].text() == "\n".join(
            ["*.plaintext", "*.decrypted", "*.ppk"]
        )
        assert operations["shared"]["credentials-file.json"].isfile()
        assert operations["credentials-env"]["default"]["subdir"][
            "credentials-env-sub-file.json"
        ].isfile()
        assert operations["credentials-env"]["default"][
            "credentials-env-file.json"
        ].isfile()
        assert operations["credentials-env"]["default"][".aws-ssh-env.pem"].isfile()
        assert operations["credentials-env"]["default"][".gitignore"].text() == "test"
        # Legacy
        assert not root["aws"].exists()
        # Upgraded
        cf = root["cf"]
        assert not cf["shared"]["aws-file.json"].isfile()
        assert cf["shared"]["aws-cf-file.json"].isfile()
        assert cf["aws-env"]["default"]["aws-env-cf-file.json"].isfile()
        assert cf["aws-env"]["default"]["subdir"]["aws-env-cf-sub-file.json"].isfile()
        assert not cf["aws-env"]["default"]["aws-env-sub-file.json"].exists()

    with tempfile.TemporaryDirectory() as root:
        create_test_fs(root)
        # subprocess.call('tree -a', shell=True, cwd=tmp_dir)
        assert ct.upgrade_cmdb_repo_to_v1_1_0(root, "")
        assert ct.cleanup_cmdb_repo_to_v1_1_0(root, "")
        # subprocess.call('tree -a', shell=True, cwd=tmp_dir)
        assert_post_upgrade_structure(root)


def test_upgrade_version_1_2_0():
    def create_test_fs(root):
        root = FSNode(root)
        root["container_template.ftl"] = "container_template"
        root["container_blueprint.ftl"] = "container_blueprint"
        root["container_exists.ftl"] = "container_exists"
        root["fragment_exists.ftl"] = "fragment_exists"

        sub_dir = root["sub_dir"]
        sub_dir["container_template.ftl"] = "container_template"
        sub_dir["container_blueprint.ftl"] = "container_blueprint"
        sub_dir["container_exists.ftl"] = "container_exists"
        sub_dir["fragment_exists.ftl"] = "fragment_exists"

    def assert_post_upgrade_structure(root):
        root = FSNode(root)
        assert not root["container_template.ftl"].exists()
        assert not root["container_blueprint.ftl"].exists()
        assert root["container_exists.ftl"].text() == "container_exists"
        assert root["fragment_exists.ftl"].text() == "fragment_exists"

        sub_dir = root["sub_dir"]
        assert not sub_dir["container_template.ftl"].exists()
        assert not sub_dir["container_blueprint.ftl"].exists()
        assert sub_dir["container_exists.ftl"].text() == "container_exists"
        assert sub_dir["fragment_exists.ftl"].text() == "fragment_exists"

    with tempfile.TemporaryDirectory() as root:
        create_test_fs(root)
        assert ct.upgrade_cmdb_repo_to_v1_2_0(root, "")
        assert_post_upgrade_structure(root)


def test_upgrade_version_1_3_0():
    CMK_STACK_DATA = {
        "Stacks": [
            {
                "Outputs": [
                    {"OutputKey": "Account", "OutputValue": "AWSId_1"},
                    {"OutputKey": "Region", "OutputValue": "AWS_Region_1"},
                ]
            }
        ]
    }

    EMPTY_STACK_DATA = {
        "Stacks": [{"Outputs": [{"OutputKey": "Custom", "OutputValue": "Custom"}]}]
    }

    UPDATED_EMPTY_STACK_DATA = {
        "Stacks": [
            {
                "Outputs": [
                    {"OutputKey": "Custom", "OutputValue": "Custom"},
                    {"OutputKey": "Account", "OutputValue": "AWSId_1"},
                    {"OutputKey": "Region", "OutputValue": "us-east-1"},
                ]
            }
        ]
    }

    def create_test_fs(root):
        root = FSNode(root)

        accounts = root["accounts"]
        accounts["account"]["config"]["account.json"] = {
            "Account": {"Id": "Id_1", "AWSId": "AWSId_1"}
        }
        cf = root["infrastructure"]["cf"]
        operations = root["infrastructure"]["operations"]

        cf["seg-cmk-1-stack.json"] = CMK_STACK_DATA
        cf[
            "stacklevel-deplymentunit-10000000000-us-east-1-stack.json"
        ] = EMPTY_STACK_DATA
        cf["stacklevel-deplyment-unit-xxxx-us-east-1-stack.json"] = EMPTY_STACK_DATA
        cf["stacklevel-deplymentunit-us-east-1-stack.json"] = EMPTY_STACK_DATA

        operations[".aws-ssh-0.pem"] = "0"
        operations[".aws-ssh-1.pem"] = "1"

    def assert_post_upgrade_structure(root):
        root = FSNode(root)

        cf = root["infrastructure"]["cf"]
        operations = root["infrastructure"]["operations"]

        assert cf["seg-cmk-1-stack.json"].json() == CMK_STACK_DATA

        assert not cf["stacklevel-deplyment-unit-xxxx-us-east-1-stack.json"].exists()
        assert not cf["stacklevel-deplymentunit-us-east-1-stack.json"].exists()

        assert (
            cf["stacklevel-deplymentunit-10000000000-us-east-1-stack.json"].json()
            == UPDATED_EMPTY_STACK_DATA
        )
        assert (
            cf["stacklevel-deplyment-unit-xxxx-Id_1-us-east-1-stack.json"].json()
            == UPDATED_EMPTY_STACK_DATA
        )
        assert (
            cf["stacklevel-deplymentunit-Id_1-us-east-1-stack.json"].json()
            == UPDATED_EMPTY_STACK_DATA
        )

        assert not operations[".aws-ssh-0.pem"].exists()
        assert not operations[".aws-ssh-1.pem"].exists()

        assert operations[".aws-Id_1-AWS_Region_1-ssh-0.pem"].text() == "0"
        assert operations[".aws-Id_1-AWS_Region_1-ssh-1.pem"].text() == "1"

    with tempfile.TemporaryDirectory() as root:
        create_test_fs(root)
        assert ct.upgrade_cmdb_repo_to_v1_3_0(root, "")
        assert_post_upgrade_structure(root)


def test_upgrade_version_1_3_1():
    import subprocess

    CMK_STACK_DATA = {
        "Stacks": [
            {
                "Outputs": [
                    {"OutputKey": "Account", "OutputValue": "AWSId_1"},
                    {"OutputKey": "Region", "OutputValue": "AWS_Region_1"},
                ]
            }
        ]
    }

    ACCOUNT_DATA = {"Account": {"Id": "10000000000", "AWSId": "AWSId_1"}}

    def create_test_fs(root):
        root = FSNode(root)

        accounts = root["accounts"]
        accounts["account"]["config"]["account.json"] = ACCOUNT_DATA
        cf = root["infrastructure"]["cf"]

        cf["seg-cmk-1-stack.json"] = CMK_STACK_DATA
        cf["stacklevel-deplymentunit-10000000001-us-east-1-stack.json"] = "1"
        cf["stacklevel-deplyment-unit-xxxx-us-east-1-stack.json"] = "2"
        cf["stacklevel-deplymentunit-us-east-1-stack.json"] = "3"

    def assert_post_upgrade_structure(root):
        root = FSNode(root)

        cf = root["infrastructure"]["cf"]

        assert cf["seg-cmk-1-stack.json"].json() == CMK_STACK_DATA
        assert (
            cf["stacklevel-deplymentunit-10000000001-us-east-1-stack.json"].text()
            == "1"
        )
        assert not cf["stacklevel-deplyment-unit-xxxx-us-east-1-stack.json"].exists()
        assert not cf["stacklevel-deplymentunit-us-east-1-stack.json"].exists()

        assert (
            cf["stacklevel-deplyment-unit-xxxx-10000000000-us-east-1-stack.json"].text()
            == "2"
        )
        assert (
            cf["stacklevel-deplymentunit-10000000000-us-east-1-stack.json"].text()
            == "3"
        )

    # Regular run without exceptional cases
    with tempfile.TemporaryDirectory() as root:
        create_test_fs(root)
        subprocess.call("tree -a", shell=True, cwd=root)
        assert ct.upgrade_cmdb_repo_to_v1_3_1(root, "")
        subprocess.call("tree -a", shell=True, cwd=root)
        assert_post_upgrade_structure(root)

    # Exceptional cases:
    # Files content is different, manual intervention required
    with tempfile.TemporaryDirectory() as root_dir:
        root = FSNode(root_dir)

        cf = root["infrastructure"]["cf"]
        accounts = root["accounts"]
        accounts["account"]["config"]["account.json"] = ACCOUNT_DATA
        cf["seg-cmk-1-stack.json"] = CMK_STACK_DATA
        cf["stacklevel-deplymentunit-us-east-1-stack.json"] = "1"
        cf["stacklevel-deplymentunit-10000000000-us-east-1-stack.json"] = "2"

        assert not ct.upgrade_cmdb_repo_to_v1_3_1(root_dir, "")

        assert cf["stacklevel-deplymentunit-us-east-1-stack.json"].text() == "1"
        assert (
            cf["stacklevel-deplymentunit-10000000000-us-east-1-stack.json"].text()
            == "2"
        )

    # Files content is equal, removing legacy file
    with tempfile.TemporaryDirectory() as root_dir:
        root = FSNode(root_dir)

        cf = root["infrastructure"]["cf"]
        accounts = root["accounts"]
        accounts["account"]["config"]["account.json"] = ACCOUNT_DATA
        cf["seg-cmk-1-stack.json"] = CMK_STACK_DATA
        cf["stacklevel-deplymentunit-us-east-1-stack.json"] = "1"
        cf["stacklevel-deplymentunit-10000000000-us-east-1-stack.json"] = "1"

        assert ct.upgrade_cmdb_repo_to_v1_3_1(root_dir, "")

        assert not cf["stacklevel-deplymentunit-us-east-1-stack.json"].exists()
        assert (
            cf["stacklevel-deplymentunit-10000000000-us-east-1-stack.json"].text()
            == "1"
        )


def test_upgrade_version_2_0_0():
    with tempfile.TemporaryDirectory() as root:
        root = FSNode(root)

        product = root["product"]
        config = product["config"].mkdir()
        config["config.json"].touchfile()
        settings = config["settings"].mkdir()
        settings["build.json"].touchfile()
        settings["shared-build.json"].touchfile()
        settings["not-a-build-file.json"].touchfile()
        solutionsv2 = config["solutionsv2"].mkdir()
        solutionsv2["solution.json"].touchfile()
        infrastructure = product["infrastructure"].mkdir()
        infrastructure_cf = infrastructure["cf"].mkdir()
        infrastructure_cf["cf.json"].touchfile()
        infrastructure_cot = infrastructure["cot"].mkdir()
        infrastructure_cot["cot.json"].touchfile()
        infrastructure_operations = infrastructure["operations"].mkdir()
        infrastructure_operations["infrastructure-operations.json"].touchfile()
        state = product["state"].mkdir()
        state["state.json"].touchfile()
        operations = product["operations"].mkdir()
        operations["operations.json"].touchfile()

        assert ct.upgrade_cmdb_repo_to_v2_0_0(root.path, "")
        assert ct.cleanup_cmdb_repo_to_v2_0_0(root.path, "")

        assert config["config.json"].isfile()
        assert not config["solutionsv2"].exists()
        assert settings["not-a-build-file.json"].isfile()
        assert not settings["build.json"].exists()
        assert not settings["shared-build.json"].exists()

        assert not infrastructure_operations.exists()
        assert not infrastructure_cf.exists()
        assert not infrastructure_cot.exists()

        assert infrastructure["builds"]["settings"]["build.json"].isfile()
        assert infrastructure["builds"]["settings"]["shared-build.json"].isfile()

        assert infrastructure["solutions"]["solution.json"].isfile()

        assert state["cot"]["cot.json"].isfile()
        assert state["cf"]["cf.json"].isfile()

        assert operations["operations.json"].isfile()
        assert operations["settings"]["infrastructure-operations.json"].isfile()


def test_upgrade_version_2_0_1():
    with tempfile.TemporaryDirectory() as root:
        root = FSNode(root)
        state = root["state"].mkdir()
        cf = state["cf"].mkdir()

        cf["level-deploymentunit-aa-us-east-1-suffix.json"].touchfile()
        # definition file should be removed
        cf["defn-deploymentunit-aa-eastus-suffix.json"].touchfile()
        # level account
        cf["account-deploymentunit-us-east-2-suffix.json"].touchfile()
        # level acccount dunit s3
        cf["account-us-east-1-suffix.json"].touchfile()
        # level product dunit cmk
        cf["product-us-east-1-suffix.json"].touchfile()
        # level seg dunit cmk
        cf["seg-key-us-east-1-suffix.json"].touchfile()
        # level seg
        cf["cont-deploymentunit-us-east-1-suffix.json"].touchfile()
        # already moved
        # level product dunit cmk
        cf["cmk"]["default"]["product-us-east-9-suffix.json"].touchfile()
        assert ct.upgrade_cmdb_repo_to_v2_0_1(root.path, "")

        assert not cf["defn-deploymentunit-aa-eastus-suffix.json"].exists()
        # global if region == 'us-east-1' otherwise default
        cmk = cf["cmk"]
        assert cmk["global"]["product-us-east-1-suffix.json"].isfile()
        assert cmk["global"]["seg-key-us-east-1-suffix.json"].isfile()

        deploymentunit = cf["deploymentunit"]
        assert deploymentunit["default"][
            "account-deploymentunit-us-east-2-suffix.json"
        ].isfile()
        assert deploymentunit["global"][
            "cont-deploymentunit-us-east-1-suffix.json"
        ].isfile()
        assert deploymentunit["global"][
            "level-deploymentunit-aa-us-east-1-suffix.json"
        ].isfile()

        s3 = cf["s3"]
        assert s3["global"]["account-us-east-1-suffix.json"].isfile()

        assert cf["cmk"]["default"]["product-us-east-9-suffix.json"].isfile()


# pretty basic test, just to check that there are no critical errors
@mock.patch("hamlet.backend.create.template.context_tree.upgrade_cmdb_repo_to_v1_0_0")
@mock.patch("hamlet.backend.create.template.context_tree.upgrade_cmdb_repo_to_v1_1_0")
@mock.patch("hamlet.backend.create.template.context_tree.upgrade_cmdb_repo_to_v1_2_0")
@mock.patch("hamlet.backend.create.template.context_tree.upgrade_cmdb_repo_to_v1_3_0")
@mock.patch("hamlet.backend.create.template.context_tree.upgrade_cmdb_repo_to_v1_3_1")
@mock.patch("hamlet.backend.create.template.context_tree.upgrade_cmdb_repo_to_v1_3_2")
@mock.patch("hamlet.backend.create.template.context_tree.upgrade_cmdb_repo_to_v2_0_0")
@mock.patch("hamlet.backend.create.template.context_tree.upgrade_cmdb_repo_to_v2_0_1")
@mock.patch("hamlet.backend.create.template.context_tree.cleanup_cmdb_repo_to_v1_0_0")
@mock.patch("hamlet.backend.create.template.context_tree.cleanup_cmdb_repo_to_v1_1_0")
@mock.patch("hamlet.backend.create.template.context_tree.cleanup_cmdb_repo_to_v1_1_1")
@mock.patch("hamlet.backend.create.template.context_tree.cleanup_cmdb_repo_to_v2_0_0")
def test_upgrade_cmdb(
    cleanup_cmdb_repo_to_v2_0_0,
    cleanup_cmdb_repo_to_v1_1_1,
    cleanup_cmdb_repo_to_v1_1_0,
    cleanup_cmdb_repo_to_v1_0_0,
    upgrade_cmdb_repo_to_v2_0_1,
    upgrade_cmdb_repo_to_v2_0_0,
    upgrade_cmdb_repo_to_v1_3_2,
    upgrade_cmdb_repo_to_v1_3_1,
    upgrade_cmdb_repo_to_v1_3_0,
    upgrade_cmdb_repo_to_v1_2_0,
    upgrade_cmdb_repo_to_v1_1_0,
    upgrade_cmdb_repo_to_v1_0_0,
):

    with tempfile.TemporaryDirectory() as root:
        root = FSNode(root)
        root[".git"].mkdir()
        ct.upgrade_cmdb(root.path, "", "", "v2.0.1")
        ct.cleanup_cmdb(root.path, "", "", "v2.0.0")
        assert root[".cmdb"].json() == {
            "Version": {"Upgrade": "v2.0.1", "Cleanup": "v2.0.0"}
        }

    cleanup_cmdb_repo_to_v2_0_0.assert_called_once()
    cleanup_cmdb_repo_to_v1_1_1.assert_called_once()
    cleanup_cmdb_repo_to_v1_1_0.assert_called_once()
    cleanup_cmdb_repo_to_v1_0_0.assert_called_once()

    upgrade_cmdb_repo_to_v2_0_1.assert_called_once()
    upgrade_cmdb_repo_to_v2_0_0.assert_called_once()
    upgrade_cmdb_repo_to_v1_3_2.assert_called_once()
    upgrade_cmdb_repo_to_v1_3_1.assert_called_once()
    upgrade_cmdb_repo_to_v1_3_0.assert_called_once()
    upgrade_cmdb_repo_to_v1_2_0.assert_called_once()
    upgrade_cmdb_repo_to_v1_1_0.assert_called_once()
    upgrade_cmdb_repo_to_v1_0_0.assert_called_once()



import click
from hamlet.backend.run import expo_app_publish as run_expo_app_publish_backend
from hamlet.command.common import exceptions
from hamlet.command.common.config import pass_options


@click.command(
    "expo-app-publish",
    short_help="Run a task based build of an Expo app binary",
    context_settings=dict(max_content_width=240),
)
@click.option(
    "-u",
    "--deployment-unit",
    required=True,
    help="mobile app deployment unit",
    envvar="DEPLOYMENT_UNIT",
)
@click.option(
    "-s",
    "--run-setup",
    is_flag=True,
    help="run setup installation to prepare",
    envvar="RUN_SETUP",
)
@click.option(
    "-f",
    "--force-binary-build",
    help="force the build of binary images",
    is_flag=True,
    envvar="FORCE_BINARY_BUILD",
)
@click.option(
    "-m",
    "--submit-binary",
    help="submit the binary for testing",
    is_flag=True,
    envvar="SUBMIT_BINARY",
)
@click.option(
    "-b",
    "--binary-build-process",
    help="sets the build process to create the binary",
    type=click.Choice(["fastlane", "turtle"]),
    envvar="BINARY_BUILD_PROCESS",
)
@click.option(
    "-l",
    "--build-logs",
    help="Show the full build logs",
    is_flag=True,
    envvar="BUILD_LOGS",
)
@click.option(
    "-e",
    "--environment-badge",
    help="Add a badge to the app icons for the environment",
    is_flag=True,
    envvar="ENVIRONMENT_BADGE",
)
@click.option(
    "-d",
    "--environment-badge-content",
    help="An override to the environment badge content",
    envvar="ENVIRONMENT_BADGE_CONTENT",
)
@click.option(
    "-n",
    "--node-package-manager",
    help="The node package manager to use for builds",
    type=click.Choice(["yarn", "npm"]),
    envvar="NODE_PACKAGE_MANAGER",
)
@click.option(
    "-v",
    "--app-version-source",
    help="The method to use to source the app version",
    type=click.Choice(["cmdb", "manifest"]),
    envvar="APP_VERSION_SOURCE",
)
@exceptions.backend_handler()
@pass_options
def expo_app_publish(options, **kwargs):
    """
    Run a task based build of an Expo app binary
    """

    args = {**options.opts, **kwargs}
    run_expo_app_publish_backend.run(**args, _is_cli=True)



import typing

from abc import ABC
from .engine import Engine


class EngineLoader(ABC):
    """
    The engine loader locates and seeds the engines available to an engine store
    """

    def __init__(self):
        pass

    def load(self) -> typing.Iterable[Engine]:
        """
        Load should locate and build instances of Engines and yield each engine
        """
        for engine in self._engines:
            yield engine



from hamlet.command import root as cli

from hamlet.command.component.describe_occurrence import (
    describe_occurrence as describe_occurrence_group,
)
from hamlet.command.component.list_occurrences import (
    list_occurrences as list_occurrences_command,
)
from hamlet.command.component.query_occurrences import (
    query_occurrences as query_occurrences_command,
)
from hamlet.command.component.component_types import (
    list_component_types as list_component_types_command,
    describe_component_type as describe_component_type_command,
    query_component_types as query_component_types_command,
)


@cli.group("component", context_settings=dict(max_content_width=240))
def component_group():
    """
    Provides information on the components used in a hamlet
    """
    pass


component_group.add_command(describe_occurrence_group)
component_group.add_command(list_occurrences_command)
component_group.add_command(query_occurrences_command)
component_group.add_command(list_component_types_command)
component_group.add_command(describe_component_type_command)
component_group.add_command(query_component_types_command)



import os
import click
from hamlet.backend.test import run as test_run_backend
from hamlet.command.common.exceptions import CommandError


@click.command(
    "run",
    context_settings=dict(max_content_width=240),
    short_help="Run pytest on specified files",
)
@click.option(
    "-t",
    "--test",
    "tests",
    multiple=True,
    help="file or directory containing tests",
    type=click.Path(dir_okay=True, file_okay=True, exists=True),
)
@click.option(
    "-p", "--pytest-args", "pytest_args", help="additional arguments for pytest"
)
@click.option("-s", "--silent", "silent", is_flag=True, help="minimize pytest output")
def run(tests, pytest_args, silent):
    """
    Discover and run tests in specified files or/and directories. If no tests paths provided
    current directory used as tests discovery root.
    """
    if not tests:
        tests = (os.getcwd(),)

    try:
        test_run_backend.run(
            testpaths=tests,
            pytest_args=pytest_args,
            silent=silent,
        )
    except Exception as e:
        raise CommandError(str(e))



import click
import os
import tempfile

from hamlet.command.common import decorators
from hamlet.command.common.exceptions import (
    backend_handler,
    HamletHomeDirUnavailableException,
)
from hamlet.env import HAMLET_GLOBAL_CONFIG
from hamlet.command.common.engine_setup import (
    setup_global_engine,
    get_engine_env,
)

try:
    from hamlet.__version__ import version
except ImportError:
    version = "unknown"


@click.group("root", context_settings=dict(max_content_width=240))
@click.version_option(version)
@decorators.common_cli_config_options
@decorators.common_generation_options
@decorators.common_logging_options
@decorators.common_engine_options
@decorators.common_district_options
@click.pass_context
@backend_handler()
def root(ctx, opts):
    """
    hamlet deploy
    """

    try:
        os.makedirs(HAMLET_GLOBAL_CONFIG.home_dir, exist_ok=True)
        testfile = tempfile.TemporaryFile(dir=HAMLET_GLOBAL_CONFIG.home_dir)
        testfile.close()

    except OSError as e:
        raise HamletHomeDirUnavailableException(str(e))

    setup_global_engine(opts.engine)
    get_engine_env(opts.engine)



from hamlet.backend.automation_tasks.base import AutomationRunner
from hamlet.backend.automation import (
    set_automation_context,
    construct_tree,
    save_cmdb_repos,
)


class SaveCMDBAutomationRunner(AutomationRunner):
    def __init__(
        self,
        account_repos,
        commit_message,
        product_repos,
        reference,
        tag,
        defer_push,
        **kwargs
    ):
        super().__init__()

        self._context_env = {
            **kwargs,
        }

        self._script_list = [
            {
                "func": construct_tree.run,
                "args": {
                    "use_existing_tree": True,
                    "_is_cli": True,
                },
            },
            {
                "func": set_automation_context.run,
                "args": {"_is_cli": True, "context_credentials": False},
            },
            {
                "func": save_cmdb_repos.run,
                "args": {
                    "account_repos": account_repos,
                    "commit_message": commit_message,
                    "product_repos": product_repos,
                    "reference": reference,
                    "defer_push": defer_push,
                    "tag": tag,
                    "_is_cli": True,
                },
            },
        ]



from hamlet.backend.common import runner


def run(
    exclude_account_dirs=None,
    product_config_reference=None,
    product_infrastructure_reference=None,
    exclude_product_dirs=None,
    account_config_reference=None,
    account_infrastructure_reference=None,
    use_existing_tree=None,
    log_level=None,
    tenant=None,
    account=None,
    product=None,
    environment=None,
    segment=None,
    _is_cli=False,
    env={},
):
    opts = {
        "-a": exclude_account_dirs,
        "-c": product_config_reference,
        "-e": use_existing_tree,
        "-i": product_infrastructure_reference,
        "-r": exclude_product_dirs,
        "-x": account_config_reference,
        "-y": account_infrastructure_reference,
    }
    env = {
        "AUTOMATION_LOG_LEVEL": log_level,
        "TENANT": tenant,
        "ACCOUNT": account,
        "PRODUCT": product,
        "ENVIRONMENT": environment,
        "SEGMENT": segment,
        **env,
    }
    runner.run(
        "constructTree.sh",
        [],
        opts,
        env,
        _is_cli,
        script_base_path_env="AUTOMATION_BASE_DIR",
    )



import os
import json
import shutil
import tempfile
import pytest
from hamlet.backend.common.exceptions import BackendException
from hamlet.backend.test.generate import run as generate_test_backend
from .conftest import DATA_DIR


def test():
    with tempfile.TemporaryDirectory() as temp_dir:
        # testing single file with output file option and without
        output_filename = os.path.join(temp_dir, "test.py")
        casefilename = os.path.join(DATA_DIR, "testcase", "secure-testcase.json")
        generate_test_backend([casefilename], output_filename)
        with open(output_filename, "rt") as f:
            file_text = f.read()
        stdout_text = generate_test_backend([casefilename], None)
        assert file_text == stdout_text
        assert "test_secure():" in file_text
        casefiles = [
            "secure-testcase.json",
            "insecure-testcase.json",
            "structure-good-testcase.json",
            "structure-bad-testcase.json",
            "invalid-syntax-testcase.json",
            "valid-syntax-testcase.json",
        ]
        # testing multiple files with output file option and without
        casefiles = list(
            map(lambda n: os.path.join(DATA_DIR, "testcase", n), casefiles)
        )
        generate_test_backend(casefiles, output_filename)
        with open(output_filename, "rt") as f:
            file_text = f.read()
        stdout_text = generate_test_backend(casefiles, None)
        assert file_text == stdout_text
        assert "test_secure():" in file_text
        assert "test_insecure():" in file_text
        assert "test_structureGood():" in file_text
        assert "test_structureBad():" in file_text
        assert "test_validSyntax():" in file_text
        assert "test_invalidSyntax():" in file_text

    # testing dir scan generation
    with tempfile.TemporaryDirectory() as temp_dir:
        copied_casefiles = []
        for filename in casefiles:
            filename = shutil.copy2(
                filename, os.path.join(temp_dir, os.path.basename(filename))
            )
            copied_casefiles.append(filename)
        assert generate_test_backend(copied_casefiles, None) == generate_test_backend(
            directory=temp_dir, output=None
        )

    # testing empty no testcase files error
    with tempfile.TemporaryDirectory() as temp_dir:
        with pytest.raises(BackendException) as einfo:
            generate_test_backend(directory=temp_dir, output=None)
        assert str(einfo.value) == "No testcase files found!"
        empty_testcase_file_path = os.path.join(temp_dir, "empty-testcase.json")
        with open(empty_testcase_file_path, "wt+") as f:
            json.dump({}, f)

        # testing empty testcase files error
        with pytest.raises(BackendException) as einfo:
            generate_test_backend(directory=temp_dir, output=None)
        assert str(einfo.value) == "No testcases found!"

        # testing extension check
        invalid_extension_filepath = os.path.join(temp_dir, "invalid.testcase.json")
        shutil.copy2(
            os.path.join(DATA_DIR, "testcase", "secure-testcase.json"),
            invalid_extension_filepath,
        )
        with pytest.raises(BackendException) as einfo:
            generate_test_backend(
                filenames=[invalid_extension_filepath], directory=None, output=None
            )
        assert (
            str(einfo.value)
            == f"Invalid extension for {invalid_extension_filepath}. Must be -testcase.json"
        )



import collections
from unittest import mock
from click.testing import CliRunner
from hamlet.command.manage.credentials_crypto import (
    credentials_crypto as manage_credentials_crypto,
)
from tests.unit.command.test_option_generation import (
    run_options_test,
    run_validatable_option_test,
)


ALL_VALID_OPTIONS = collections.OrderedDict()
ALL_VALID_OPTIONS["!-n,--credential-path"] = "credential_path"
ALL_VALID_OPTIONS["!-y,--credential-type"] = ["login", "api", "env"]
ALL_VALID_OPTIONS["-e,--credential-email"] = "email@email.com"
ALL_VALID_OPTIONS["-f,--crypto-file"] = "crypto_file"
ALL_VALID_OPTIONS["-i,--credential-id"] = "credential_id"
ALL_VALID_OPTIONS["-s,--credential-secret"] = "credential_secret"
ALL_VALID_OPTIONS["-v,--visible"] = [True, False]


@mock.patch(
    "hamlet.command.manage.credentials_crypto.manage_credentials_crypto_backend"
)
def test_input_valid(manage_credentials_crypto_backend):
    runner = CliRunner()
    with runner.isolated_filesystem():
        open("crypto_file", "w").close()
        # testing that's impossible to run without full set of required options
        run_options_test(
            runner,
            manage_credentials_crypto,
            ALL_VALID_OPTIONS,
            manage_credentials_crypto_backend.run,
        )


@mock.patch(
    "hamlet.command.manage.credentials_crypto.manage_credentials_crypto_backend"
)
def test_input_validation(manage_credentials_crypto_backend):
    runner = CliRunner()
    with runner.isolated_filesystem():
        open("crypto_file", "w").close()
        run_validatable_option_test(
            runner,
            manage_credentials_crypto,
            manage_credentials_crypto_backend.run,
            {"-n": "credential_path", "-y": "login"},
            [("-y", "invalidtype", "login"), ("-f", "nonexistingfile", "crypto_file")],
        )



from ..loader import loader


def prepare_cf_testcase_context(case):
    json_structure = case.get("json_structure", {})
    match = json_structure.get("match", [])
    stringified_match = []
    for match_case in match:
        value = match_case["value"]
        if isinstance(value, str):
            value = value.replace('"', r"\"")
            value = value.replace("'", r"\'")
            value = f'"{value}"'
        stringified_match.append({"path": match_case["path"], "value": value})

    if json_structure:
        if match:
            json_structure["match"] = stringified_match


def prepare_context(cases):
    cfn_lint_test = False
    cfn_structure_test = False
    json_structure_test = False
    checkov_test = False
    prepared_cases = list()
    ordered_casenames = list(cases.keys())
    ordered_casenames.sort()
    for casename in ordered_casenames:
        data = cases[casename]
        type = None
        if data["filename"].endswith(".json"):
            type = "cf"
        if type == "cf":
            prepare_cf_testcase_context(data)
            cfn_lint_test = cfn_lint_test or "cfn_lint" in data
            checkov_test = checkov_test or "checkov" in data
            cfn_structure_test = cfn_structure_test or "cfn_structure" in data
            json_structure_test = json_structure_test or "json_structure" in data
        prepared_cases.append((casename, data, type))
    return {
        "checkov_test": checkov_test,
        "cfn_lint_test": cfn_lint_test,
        "json_validation_test": cfn_structure_test or json_structure_test,
        "cfn_structure_test": cfn_structure_test,
        "json_structure_test": json_structure_test,
        "cases": prepared_cases,
    }


def render(cases=None):
    context = prepare_context(cases)
    template = loader.get_template("test_template_base.py.tmpl")
    return template.render(**context)



from hamlet.backend.common import runner


def run(
    delete=None,
    deployment_initiate=None,
    deployment_group=None,
    deployment_monitor=None,
    region=None,
    deployment_scope=None,
    deployment_unit=None,
    deployment_wait=None,
    deployment_unit_subset=None,
    output_dir=None,
    dryrun=None,
    log_level=None,
    root_dir=None,
    tenant=None,
    account=None,
    product=None,
    environment=None,
    segment=None,
    _is_cli=False,
    **kwargs,
):
    options = {
        "-d": delete,
        "-i": deployment_initiate,
        "-m": deployment_monitor,
        "-l": deployment_group,
        "-r": region,
        "-s": deployment_scope,
        "-u": deployment_unit,
        "-w": deployment_wait,
        "-y": dryrun,
        "-z": deployment_unit_subset,
        "-o": output_dir,
    }
    env = {
        "GENERATION_LOG_LEVEL": log_level,
        "ROOT_DIR": root_dir,
        "TENANT": tenant,
        "ACCOUNT": account,
        "PRODUCT": product,
        "ENVIRONMENT": environment,
        "SEGMENT": segment,
    }
    runner.run("manageDeployment.sh", [], options, env, _is_cli)



import os
import tempfile
from abc import ABC

from configparser import ConfigParser


class AutomationRunner(ABC):
    """
    Executes an automation based task
    """

    def __init__(self, **kwargs):
        self._context_env = kwargs
        self._script_list = []

    @staticmethod
    def _load_properties_to_context(properties_file):

        itemDict = {}

        if os.path.isfile(properties_file):
            config = ConfigParser(strict=False)

            s_config = open(properties_file, "r").read()
            s_config = f"[ini]\n{s_config}"

            config.read_string(s_config)
            items = config.items("ini")
            for key, value in items:
                itemDict[key] = value

        return itemDict

    def run(self):

        self._context_env["AUTOMATION_PROVIDER"] = "hamletcli"

        with tempfile.TemporaryDirectory() as tmp_dir:

            self._context_env["AUTOMATION_DATA_DIR"] = tmp_dir

            for script in self._script_list:

                result = script["func"](env=self._context_env, **script["args"])

                if isinstance(result, dict):
                    self._context_env.update(result)

                script_context_envs = self._load_properties_to_context(
                    os.path.join(tmp_dir, "context.properties")
                )
                self._context_env = {**self._context_env, **script_context_envs}



from abc import ABC


class EnginePartInterface(ABC):
    """
    An engine part defines a functional component of the hamlet tooling
    Each part has a fixed type with instances have a link to a source that
    has the component tooling
    """

    type = None
    description = ""

    def __init__(self, source_name, source_path):
        self.source_path = source_path
        self.source_name = source_name


class CoreEnginePart(EnginePartInterface):
    type = "core-engine"
    description = "core engine source"


class AWSEnginePluginPart(EnginePartInterface):
    type = "engine-plugin-aws"
    description = "default aws engine plugin"


class AzureEnginePluginPart(EnginePartInterface):
    type = "engine-plugin-azure"
    description = "default azure engine plugin"


class CMDBEnginePluginPart(EnginePartInterface):
    type = "engine-plugin-cmdb"
    description = "cmdb input engine plugin"


class WrapperEnginePart(EnginePartInterface):
    type = "engine-wrapper"
    description = "hamlet freemarker wrapper jar file"


class BashExecutorEnginePart(EnginePartInterface):
    type = "executor-bash"
    description = "bash based executor"



import os

from hamlet.backend import query as query_backend

GET_PROPERTIES_ENVIRONMENT_QUERY = "Environment"


def get_automation_properties(**kwargs):

    query_args = {
        **kwargs,
        "generation_entrance": "releaseinfo",
        "output_filename": "releaseinfo-config.json",
        "output_format": "default",
        "use_cache": False,
    }
    automation_properties = query_backend.run(
        **query_args, cwd=os.getcwd(), query_text=GET_PROPERTIES_ENVIRONMENT_QUERY
    )

    return automation_properties



import json
from unittest import mock
from click.testing import CliRunner


from hamlet.command.engine import (
    list_engines,
    describe_engine,
    clean_engines,
    install_engine,
    set_engine,
    env,
)


def mock_engine(name, description, installed, digest="", hidden=False):
    mock_engine = mock.Mock()
    mock_engine.name = name
    mock_engine.description = description
    mock_engine.digest = digest
    mock_engine.hidden = hidden
    type(mock_engine).installed = mock.PropertyMock(return_value=installed)

    return mock_engine


def mock_backend():
    def decorator(func):
        @mock.patch("hamlet.command.engine.engine_store")
        def wrapper(mock_engine_store, *args, **kwargs):

            mock_engines = [
                mock_engine(
                    name="Name[1]",
                    description="Description[1]",
                    installed=False,
                    hidden=False,
                ),
                mock_engine(
                    name="Name[2]",
                    description="Description[2]",
                    installed=True,
                    digest="Digest[2]",
                    hidden=True,
                ),
            ]

            mock_engine_store.get_engines.return_value = mock_engines
            mock_engine_store.find_engines.return_value = mock_engines
            type(mock_engine_store).global_engine = mock.PropertyMock(
                return_value="Name[2]"
            )

            return func(mock_engine_store, *args, **kwargs)

        return wrapper

    return decorator


@mock_backend()
def test_list_engines(mock_engine_store):
    cli = CliRunner()
    result = cli.invoke(list_engines, ["--output-format", "json"])
    print(result.output)
    assert result.exit_code == 0
    result = json.loads(result.output)
    print(result)
    assert len(result) == 1
    assert {
        "name": "Name[1]",
        "description": "Description[1]",
        "digest": "",
        "installed": False,
        "global": False,
        "update_available": None,
    } in result


@mock.patch("json.dumps", mock.MagicMock(return_value="{}"))
@mock_backend()
def test_describe_engine(mock_engine_store):
    cli = CliRunner()
    result = cli.invoke(describe_engine)
    print(result.output)
    assert result.exit_code == 0


@mock_backend()
def test_install_engine(mock_engine_store):
    cli = CliRunner()
    result = cli.invoke(install_engine, ["engine[1]"])
    print(result.exception)
    assert result.exit_code == 0


@mock_backend()
def test_env(mock_engine_store):
    cli = CliRunner()
    result = cli.invoke(env, [])
    print(result.exception)
    assert result.exit_code == 0
    print(result.output)
    assert result.output == "# run eval $(hamlet engine env) to set variables\n"


@mock_backend()
def test_set_engine(mock_engine_store):
    cli = CliRunner()
    result = cli.invoke(set_engine, ["Name[1]"])
    print(result.exception)
    assert result.exit_code == 0


@mock_backend()
def test_clean_engines(mock_engine_store):
    cli = CliRunner()
    result = cli.invoke(clean_engines, [])
    print(result.exception)
    assert result.exit_code == 0



from marshmallow import Schema, fields


class CFNStructure(Schema):
    class Resource(Schema):
        id = fields.Str(required=True)
        type = fields.Str(required=True)

    class Output(Schema):
        id = fields.Str(required=True)

    resource = fields.List(fields.Nested(Resource))
    output = fields.List(fields.Nested(Output))


class JSONStructure(Schema):
    class Match(Schema):
        path = fields.Str(required=True)
        value = fields.Raw(required=True)

    class Exists(Schema):
        path = fields.Str(required=True)

    class Length(Schema):
        path = fields.Str(required=True)
        value = fields.Integer(required=True, strict=True)

    class NotEmpty(Schema):
        path = fields.String(required=True)

    match = fields.List(fields.Nested(Match))
    exists = fields.List(fields.Nested(Exists))
    length = fields.List(fields.Nested(Length))
    not_empty = fields.List(fields.Nested(NotEmpty))


class CheckovStructure(Schema):
    framework = fields.Str()
    skip_checks = fields.List(fields.Str())


class CFNLintStructure(Schema):
    ignore_checks = fields.List(fields.Str())


class Testcase(Schema):
    filename = fields.Str(required=True)
    cfn_lint = fields.Nested(CFNLintStructure)
    checkov = fields.Nested(CheckovStructure)
    cfn_structure = fields.Nested(CFNStructure)
    json_structure = fields.Nested(JSONStructure)
